import datetime

from ..utilities import settings, util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class Product(Request):

    """
    The Products API allows you create and manage inventories on your integration
    """
    path = '/product'

    def create(self, name: str, description: str, price: int, currency: str, unlimited: bool = None, quantity: int = None):
        """Create a product on your integration

        Args:
            name (str): Name of product
            description (str): A description for this product
            price (int): Price should be in kobo if currency is NGN, pesewas,
                if currency is GHS, and cents, if currency is ZAR
            currency (str): Currency in which price is set. Allowed values are: NGN, GHS, ZAR or USD
            unlimited (bool, optional): Set to true if the product has unlimited stock.
                Leave as false if the product has limited stock. Defaults to None.
            quantity (int, optional): Number of products in stock. Use if unlimited is false. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        payload = util.generate_payload(locals(), 'currency')
        payload['currency'] = util.check_membership(
            settings.CURRENCIES, currency, 'currency')

        if quantity and quantity != 0:
            assert unlimited is None or unlimited is False, "set unlimited to False then set quantity, set unlimited to True otherwise."
            payload['quantity'] = quantity

        if unlimited is True:
            assert quantity is None, "you can't set quantity with unlimited set to True"
            payload['unlimited'] = True

        return self.post(self.path, payload)

    def list_products(self, per_page: int = None, page: int = None, from_date: datetime.datetime | datetime.date | str = None, to_date: datetime.datetime | datetime.date | str = None):
        """List products available on your integration.

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.. Defaults to None.
            page (int, optional): Specify exactly what page you want to retrieve.
                If not specify we use a default value of 1.. Defaults to None.
            from_date (datetime.datetime | datetime.date | str, optional): A timestamp from which to start listing product
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            to_date (datetime.datetime | datetime.date | str, optional): A timestamp at which to stop listing product
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)
        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch(self, product_id: int):
        """Get details of a product on your integration.

        Args:
            product_id (int): The product ID you want to fetch

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/{product_id}"
        return self.get(path)

    def update(self, product_id: int, name: str = None, description: str = None, price: int = None, currency: str = None, unlimited: bool = None, quantity: int = None):
        """Update a product details on your integration

        Args:
            product_id (int): Product ID
            name (str, optional): Name of product. Defaults to None.
            description (str, optional): A description for this product. Defaults to None.
            price (int, optional): Price should be in kobo if currency is NGN, pesewas,
                if currency is GHS, and cents, if currency is ZAR. Defaults to None.
            currency (str, optional): Currency in which price is set.
                Allowed values are: NGN, GHS, ZAR or USD. Defaults to None.
            unlimited (bool, optional): Set to true if the product has unlimited stock.
                Leave as false if the product has limited stock.
            quantity (int, optional): Number of products in stock.
                Use if unlimited is false. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/{product_id}"
        payload = {key: value for key, value in locals(
        ).items() if key not in ('product_id', 'path') and value is not None}

        if currency:
            payload['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        if quantity and quantity != 0:
            assert unlimited is None or unlimited is False, "set unlimited to False then set quantity, set unlimited to True otherwise."
            payload['quantity'] = quantity

        if unlimited is True:
            assert quantity is None, "you can't set quantity with unlimited set to True"
            payload['unlimited'] = True

        return self.put(path, payload)



from ..utilities import settings, util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class Miscellaneous(Request):

    """The Miscellaneous API are supporting APIs that can be used to provide more details to other APIs
    """

    def list_banks(self, country: str = None, use_cursor: bool = False, per_page: int = None, next: str = None, previous: str = None, gateway: str = None, type: str = None, currency: str = None):
        """Get a list of all supported banks and their properties

        Args:
            country (str, optional): The country from which to obtain the list of supported banks.
                e.g country=ghana or country=nigeria. Defaults to None.
            use_cursor (bool, optional): Use_curFlag to enable cursor pagination on the endpointsor. Defaults to False.
            per_page (int, optional): The number of objects to return per page.
                Defaults to 50, and limited to 100 records per page.
            next (str, optional): A cursor that indicates your place in the list.
                It can be used to fetch the next page of the list. Defaults to None.
            previous (str, optional): A cursor that indicates your place in the list.
                It should be used to fetch the previous page of the list after an intial next request. Defaults to None.
            gateway (str, optional): The gateway type of the bank.
                It can be one of these: [emandate, digitalbankmandate]. Defaults to None.
            type (str, optional): Type of financial channel. For Ghanaian channels,
                please use either mobile_money for mobile money channels OR ghipps for bank channels. Defaults to None.
            currency (str, optional): Any of NGN, USD, GHS or ZAR. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = '/bank'
        params = util.generate_payload(locals(), 'per_page', 'currency').update(
            util.check_query_params(per_page=per_page))

        if currency:
            params['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        if params:
            path = util.handle_query_params(path, params)
        return self.get(path)

    def list_providers(self, pay_with_bank_transfer: bool = None):
        """Get a list of all providers for Dedicated Virtual Account

        Args:
            pay_with_bank_transfer (bool, optional): A flag to filter for available providers. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = '/bank'
        params = util.generate_payload(locals())

        if params:
            path = util.handle_query_params(path, params)

        return self.get(path)

    def list_search_countries(self):
        """Gets a list of Countries that Paystack currently supports

        Returns:
            JSON: Data fetched from API
        """

        path = '/country'

        return self.get(path)

    def list_states(self, country: int = None):
        """Get a list of states for a country for address verification.

        Args:
            country (int, optional): The country code of the states to list.
                It is gotten after the charge request. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """
        path = f'/address_verification/states?country={country}' if country else '/address_verification/states'

        return self.get(path)



from datetime import date, datetime

from ..utilities import settings, util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class Disputes(Request):

    """
    The Disputes API allows you manage transaction disputes on your integration
    """

    path = '/dispute'

    def list_disputes(self, page: int = None, per_page: int = None, from_date: date | datetime | str = None, to_date: date | datetime | str = None, transaction_id: int = None, status: str = None):
        """List all the dispute filed against you.

        Args:
            page (int, optional): Specify exactly what dispute you want to page. If not specify we use a default value of 1.
                Defaults to None.
            per_page (int, optional): Specify how many records you want to retrieve per page. If not specify we use a default value of 50.
                Defaults to None.
            from_date (date | datetime | str, optional): A timestamp from which to start listing dispute e.g. 2016-09-21.
                Defaults to None.
            to_date (date | datetime | str, optional): A timestamp at which to stop listing dispute e.g. 2016-09-21.
                Defaults to None.
            transaction_id (int, optional): Transaction Id. Defaults to None.
            status (str, optional): Dispute Status. Acceptable values: { awaiting-merchant-feedback | awaiting-bank-feedback | pending | resolved }.
                Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)

        if transaction_id:
            params['transaction'] = transaction_id

        if status:
            params['status'] = util.check_membership(
                settings.DISPUTE_STATUSES, status, 'status')

        if params:
            path = util.handle_query_params(self.path, params)
            return self.get(path)
        return self.get(self.path)

    def fetch(self, dispute_id: int):
        """Get more details about a dispute

        Args:
            dispute_id (int): The ID of the dispute your want to fetch

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{dispute_id}'
        return self.get(path)

    def list_transaction_disputes(self, transaction_id: int):
        """Retrieve disputes for a particular transaction

        Args:
            transaction_id (int): ID of the transaction whose dispute you want to fetch

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/transaction/{transaction_id}'
        return self.get(path)

    def update(self, dispute_id: int, refund_amount: int = None, uploaded_filename: str = None):
        """Update details of a dispute on your integration

        Args:
            dispute_id (int): The ID of the dispute to be updated
            refund_amount (int, optional): The amount to refund, in kobo if currency is NGN, pesewas,
                if currency is GHS, and cents, if currency is ZAR. Defaults to None
            uploaded_filename (str, optional): filename of attachment returned via response from upload url(GET /dispute/:id/upload_url).
                Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{dispute_id}'
        payload = util.generate_payload(locals(), 'dispute_id')
        return self.put(path, payload)

    def add_evidence(self, dispute_id: int, customer_email: str, customer_name: str, customer_phone: str, service_details: str, delivery_address: str = None, delivery_date: date | datetime | str = None):
        """Provide evidence for the dispute.

        Args:
            dispute_id (int): ID of the dispute to add evidence for.
            customer_email (str): Customer's email address
            customer_name (str): Customer's name
            customer_phone (str):Customer's phone number
            service_details (str): Details of the service involved
            delivery_address (str, optional): Delivery Address. Defaults to None.
            delivery_date (date | datetime | str, optional): ISO 8601 representation of delivery date (YYYY-MM-DD). 
                Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{dispute_id}/evidence'
        payload = util.generate_payload(locals(), 'dispute_id')
        payload['customer_email'] = util.check_email(customer_email)

        if delivery_date:
            payload['delivery_date'] = util.handle_date(delivery_date)
        return self.post(path, payload)

    def get_upload_url(self, dispute_id: int, upload_filename: str = None):
        """Get URL to upload a dispute evidence.

        Args:
            dispute_id (int): ID of the dispute
            upload_filename (str, optional): The file name, with its extension, that you want to upload.
                e.g filename.pdf. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{dispute_id}/upload_url'
        if upload_filename:
            path = util.handle_query_params(
                path, {'upload_filename': upload_filename})
        return self.get(path)

    def resolve(self, dispute_id: int, resolution: str, message: str, refund_amount: int, upload_url: str, evidence_id: int = None):
        """Resolve disputes on your integration

        Args:
            dispute_id (int): ID of the dispute to be resolved.
            resolution (str): Dispute resolution. Accepted values: { merchant-accepted | declined }.
            message (str): Reason for resolving
            refund_amount (int): The amount to refund, in kobo if currency is NGN, pesewas, if currency is GHS, and cents, if currency is ZAR
            upload_url (str): Filename of attachment returned via response from upload url(GET /dispute/:id/upload_url)
            evidence_id (int, optional): Evidence Id for fraud claims. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{dispute_id}/resolve'
        payload = util.generate_payload(locals(), 'dispute_id')
        payload['resolution'] = util.check_membership(
            settings.DISPUTE_RESOLUTION, resolution, 'resolution')

        if evidence_id:
            payload['evidence'] = payload.pop('evidence_id')
        return self.put(path, payload)

    def export(self, per_page: int = None, page: int = None, from_date: date | datetime | str = None, to_date:  date | datetime | str = None, transaction_id: int = None, status: str = None):
        """Export disputes available on your integration\

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.. Defaults to None.
            page (int, optional): Specify exactly what dispute you want to page.
                If not specify we use a default value of 1.. Defaults to None.
            from_date (date | datetime | str, optional): A timestamp from which to start listing dispute
                e.g. 2016-09-21. Defaults to None.
            to_date (date | datetime | str, optional): A timestamp at which to stop listing dispute
                e.g. 2016-09-21. Defaults to None.
            transaction_id (int, optional): Transaction ID. Defaults to None.
            status (str, optional): Dispute Status. Acceptable values: { awaiting-merchant-feedback | awaiting-bank-feedback | pending | resolved }. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/export'
        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)

        if transaction_id:
            params['transaction'] = transaction_id

        if status:
            params['status'] = util.check_membership(
                settings.DISPUTE_STATUSES, status, 'status')

        if params:
            path = util.handle_query_params(path, params)
        return self.get(path)



from datetime import date, datetime

from ..utilities import settings, util, decorators
from ..utilities.errors import UnwantedArgumentsError
from ..utilities.request import Request


@decorators.class_type_checker
class Invoice(Request):

    """
    The Invoices API allows you issue out and manage payment requests
    """

    path = '/paymentrequest'

    def create(self, customer: int | str, description: str, due_date: date | datetime | str, amount: int = None, line_items: list[dict[str]] = None, tax: list[dict[str]] = None, currency: str = None, send_notification: bool = None, draft: bool = None, has_invoice: bool = None, invoice_number: int = None, split_code: str = None):
        """Create an invoice for payment on your integration.

        Args:
            customer (int | str): Customer ID or code.
            description (str): A short description of the payment request
            due_date (date | datetime | str): ISO 8601 representation of request due date
                eg. 2016-09-24T00:00:05.000Z, 2016-09-21. 
            amount (int, optional): Payment request amount. It should be used when line items and tax values aren't specified.
                Defaults to None.
            line_items (list[dict[str]], optional): Array of line items int the format [{"name":"item 1", "amount":2000, "quantity": 1}]. Defaults to None.
            tax (list[dict[str]], optional): Array of taxes to be charged in the format [{"name":"VAT", "amount":2000}]. Defaults to None.
            currency (str, optional): Specify the currency of the invoice. Allowed values are NGN, GHS, ZAR and USD. Defaults to NGN. Defaults to None.
            send_notification (bool, optional): Indicates whether Paystack sends an email notification to customer. Defaults to true. Defaults to None.
            draft (bool, optional): Indicate if request should be saved as draft. Defaults to false and overrides send_notification. Defaults to None.
            has_invoice (bool, optional): Set to true to create a draft invoice (adds an auto incrementing invoice number if none is provided) even if there are no line_items or tax passed. Defaults to None.
            invoice_number (int, optional): Numeric value of invoice. Invoice will start from 1 and auto increment from there. This field is to help override whatever value Paystack decides. Auto increment for subsequent invoices continue from this point.. Defaults to None.
            split_code (str, optional): The split code of the transaction split. e.g. SPL_98WF13Eb3w. Defaults to None.

        Raises:
            UnwantedArgumentsError: Raised when an argument passed is unwanted.

        Returns:
            JSON: Data fetched from API
        """

        if (amount and (line_items or tax)) or not (amount or line_items):
            raise UnwantedArgumentsError(
                'provide either amount or line_items not both!')

        payload = util.generate_payload(
            locals(), 'customer_id', 'customer_code', 'due_date', 'currency', 'split_code')
        payload['customer'] = util.id_or_code(customer, data=settings.CUSTOMER)
        payload['due_date'] = util.handle_date(due_date)

        if currency:
            payload['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        if split_code:
            payload['split_code'] = util.check_code(
                settings.SPLIT, split_code)

        return self.post(self.path, payload)

    def list_invoices(self, per_page: int = None, page: int = None, customer: int = None, status: str = None, currency: str = None, from_date: date | datetime | str = None, to_date: date | datetime | str = None, include_archive: str = None):
        """List invoice available on your integration

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.
            page (int, optional): Specify exactly what invoice you want to page.
                If not specify we use a default value of 1.
            customer (int, optional): Filter by customer ID. Defaults to None.
            currency (str, optional): Filter by currency. Allowed values are NGN, GHS, ZAR and USD.
            status (str, optional): Filter by invoice status. Defaults to None.
            from_date (date | datetime | str, optional): A timestamp from which to start listing invoice
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            to_date (date | datetime | str, optional): A timestamp at which to stop listing invoice
                e.g. 2016-09-24T00:00:05.000Z, 201. Defaults to None.
            include_archive (str, optional): Show archived invoices. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.generate_payload(
            locals(), 'per_page', 'page', 'from_date', 'to_date', 'currency')
        params.update(
            util.check_query_params(per_page=per_page, page=page, from_date=from_date, to_date=to_date))

        if currency:
            params['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def view(self, invoice: int | str):
        """Get details of an invoice on your integration.

        Args:
            invoice (int | str): ID or code of the invoice

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{util.id_or_code(invoice, data=settings.INVOICE)}'

        return self.get(path)

    def verify(self, invoice_code: str):
        """Verify details of an invoice on your integration.

        Args:
            invoice_code (str): The invoice code.

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/verify/{util.check_code(settings.INVOICE, invoice_code)}"
        return self.get(path)

    def send_notification(self, invoice_code: str):
        """Send notification of an invoice to your customers

        Args:
            invoice_code (str): The invoice code

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/notify/{util.check_code(settings.INVOICE, invoice_code)}"
        return self.get(path)

    def total(self):
        """Get invoice metrics for dashboard

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/totals"
        return self.get(path)

    def finalize(self, invoice_code: str):
        """Finalize a Draft Invoice

        Args:
            invoice_code (str): The invoice code

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/finalize/{util.check_code(settings.INVOICE, invoice_code)}"
        return self.get(path)

    def update(self, invoice: int | str, customer: int | str = None, amount: int = None, currency: str = None, due_date: date | datetime | str = None, description: str = None, line_items: list[dict[str]] = None, tax: list[dict[str]] = None, send_notification: bool = None, draft: bool = None, invoice_number: int = None, split_code: str = None):
        """Update an invoice details on your integration.

        Args:
            invoice_id (int, optional): The invoice id or code.
            customer_code (str, optional): Customer's code or ID. Defaults to None.
            amount (int, optional): Payment request amount. Only useful if line items and tax values are ignored.
                endpoint will throw a friendly warning if neither is available.. Defaults to None.
            currency (str, optional): Specify the currency of the invoice. Allowed values are NGN, GHS, ZAR and USD Defaults to NGN
            due_date (date | datetime | str, optional): ISO 8601 representation of request due date. Defaults to None.
            description (str, optional): A short description of the payment request. Defaults to None.
            line_items (list[dict[str]], optional): Array of line items int the format [{"name":"item 1", "amount":2000}]. Defaults to None.
            tax (list[dict[str]], optional): Array of taxes to be charged in the format [{"name":"VAT", "amount":2000}]. Defaults to None.
            send_notification (bool, optional): Indicates whether Paystack sends an email notification to customer. Defaults to true.
            draft (bool, optional): Indicate if request should be saved as draft. Defaults to false and overrides send_notification.
            invoice_number (int, optional): Numeric value of invoice.
                Invoice will start from 1 and auto increment from there. This field is to help override whatever value Paystack decides. Auto increment for subsequent invoices continue from this point.. Defaults to None.
            split_code (str, optional): The split code of the transaction split. e.g. SPL_98WF13Eb3w. Defaults to None.

        Return:
            JSON: Data fetched from API
        """
        
        path = f'{self.path}/{util.id_or_code(invoice, data=settings.INVOICE)}'

        payload = util.generate_payload(
            locals(), 'customer_code', 'customer_id', 'invoice_code', 'invoice_id')

        payload['customer'] = util.id_or_code(customer, data=settings.CUSTOMER)

        if amount:
            assert not (
                tax or line_items), "you can either provide amount or line_items and/or tax"

        if (tax or line_items):
            assert not amount, "you can either provide amount or line_items and/or tax"

        if currency:
            payload['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        if due_date:
            payload['due_date'] = util.handle_date(due_date)

        if split_code:
            payload['split_code'] = util.check_code(
                settings.SPLIT, split_code)

        return self.put(path, payload)

    def archive(self, invoice: int | str):
        """Used to archive an invoice. Invoice will no longer be fetched on list or returned on verify:

        Args:
            invoice (int | str): The invoice ID or code

        Returns:
            JSON: Data fetched from API
        """
        path = f"{self.path}/archive/{util.id_or_code(invoice, data=settings.INVOICE)}"

        return self.get(path)



from datetime import date, datetime
from typing import Sequence

from ..utilities import settings, util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class BulkCharges(Request):

    """
    The Bulk Charges API allows you create and manage multiple recurring payments from your customers
    """

    path = '/bulkcharge'

    def initiate(self, charges: Sequence):
        """Send list of tuples with authorization codes and amount (in kobo if currency is NGN, pesewas, if currency is GHS, and cents, if currency is ZAR ) 
        so we can process transactions as a batch.

        Args:
            charges ( Sequence[tuple[str, int]] ): List of tuples each containing an authorization code as a string
                and an amount to be charged as an integer eg [( "AUTH_n95vpedf", 2500 ), ( "AUTH_n95vpedf", 2500 ), ( "AUTH_n95vpedf", 2500 )]

        Returns:
            json: Data fetched from paystack API
        """

        payload = [{'authorization': util.check_code(
            settings.AUTHORIZATION, x[0]), 'amount': x[1]} for x in charges]

        return self.post(self.path, payload)

    def list_batches(self, per_page: int = None, page: int = None, from_date: date | datetime | str = None, to_date: date | datetime | str = None):
        """This lists all bulk charge batches created by the integration. Statuses can be active, paused, or complete.

        Args:
            per_page (Optional[ int ]): Specify how many records you want to retrieve per page. If not specify we use a default value of 50.

            page (Optional[ int ]): Specify exactly what transfer you want to page. If not specify we use a default value of 1.

            from_date (Optional[ date | datetime | str ]): A timestamp from which to start listing batches e.g. 2016-09-24T00:00:05.000Z, 2016-09-21

            to_date (Optional[ date | datetime | str ]): A timestamp at which to stop listing batches e.g. 2016-09-24T00:00:05.000Z, 2016-09-21


        Returns:
            json: Data fetched from paystack API
        """

        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)
        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch_batches(self, id_or_code: int | str):
        """This endpoint retrieves a specific batch code.
        It also returns useful information on its progress by way of the total_charges and pending_charges attributes.
    
        Args:
            id_or_code (int | str): An ID or code for the charge whose batches you want to retrieve.


        Returns:
            json: Data fetched from paystack API
        """

        path = f"{self.path}/{util.id_or_code(id_or_code, data=settings.BULK_CHARGE_BATCH)}"

        return self.get(path)

    def fetch_charges_in_batch(self, id_or_code: int | str, status: str = None, per_page: int = None, page: int = None, from_date: date | datetime | str = None, to_date: date | datetime | str = None):
        """This endpoint retrieves the charges associated with a specified batch code.
        Pagination parameters are available.
        You can also filter by status. Charge statuses can be pending, success or failed.

        Args:
            id_or_code (int | str): An ID or code for the batch whose charges you want to retrieve.

            status ( str ): Either one of these values: 'pending', 'success' or 'failed'

            per_page (Optional[ int ]): Specify how many records you want to retrieve per page. If not specify we use a default value of 50.

            page (Optional[ int ]): Specify exactly what transfer you want to page. If not specify we use a default value of 1.

            from_date (Optional[ date | datetime | str ]): A timestamp from which to start listing batches e.g. 2016-09-24T00:00:05.000Z, 2016-09-21

            to_date (Optional[ date | datetime | str ]): A timestamp at which to stop listing batches e.g. 2016-09-24T00:00:05.000Z, 2016-09-21

        Returns:
            json: Data fetched from paystack API
        """

        path = f"{self.path}/{util.id_or_code(id_or_code, data=settings.BULK_CHARGE_BATCH)}/charges"

        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)
        if status:
            params['status'] = util.check_membership(
                settings.BULK_CHARGE_STATUSES, status, status)

        if params:
            return self.get(util.handle_query_params(path, params))
        return self.get(path)

    def pause_batch(self, batch_code: str):
        """Use this endpoint to pause processing a batch

        Args:
            batch_code (str): The batch code for the bulk charge you want to pause

        Returns:
            json: Data fetched from paystack API
        """

        path = f"{self.path}/pause/{util.check_code(settings.BULK_CHARGE_BATCH, batch_code)}"

        return self.get(path)

    def resume_batch(self, batch_code: str):
        """Use this endpoint to resume processing a batch

        Args:
            batch_code (str): The batch code for the bulk charge you want to resume

        Returns:
            json: Data fetched from paystack API
        """
        path = f"{self.path}/resume/{util.check_code(settings.BULK_CHARGE_BATCH, batch_code)}"

        return self.get(path)



from datetime import date, datetime

from ..utilities import util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class PaymentPages(Request):

    """
    The Payment Pages API provides a quick and secure way to collect payment for products.
    """

    path = "/page"

    def create(self, name: str, description: str, amount: int, slug: str = None, metadata: dict = None, redirect_url: str = None, custom_fields: list = None):
        """Create a payment page on your integration

        Args:
            name (str): Name of page
            description (str): A description for this page
            amount (int): Amount should be in kobo if currency is NGN, pesewas,
                if currency is GHS, and cents, if currency is ZAR
            slug (str, optional): URL slug you would like to be associated with this page.
                Page will be accessible at https://paystack.com/pay/[slug]. Defaults to None.
            metadata (dict, optional): Extra data to configure the payment page including subaccount,
                logo image, transaction charge. Defaults to None.
            redirect_url (str, optional): If you would like Paystack to redirect someplace upon successful payment,
                specify the URL here. Defaults to None.
            custom_fields (list, optional): If you would like to accept custom fields,
                specify them here. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        payload = util.generate_payload(locals())
        return self.post(self.path, payload)

    def list_pages(self, per_page: int = None, page: int = None, from_date: date | datetime | str = None, to_date: date | datetime | str = None):
        """List payment pages available on your integration.

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.
            page (int, optional): Specify exactly what page you want to retrieve.
                If not specify we use a default value of 1.
            from_date (date | datetime | str, optional): A timestamp from which to start listing page
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            to_date (date | datetime | str, optional): A timestamp at which to stop listing page
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)
        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch(self, id_or_slug: int | str):
        """Get details of a payment page on your integration.

        Args:
            id_or_slug (int | str): The page ID or slug you want to fetch

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/{id_or_slug}"
        return self.get(path)

    def update(self, id_or_slug: int | str, name: str = None, description: str = None, amount: int = None, active: bool = None):
        """Update a payment page details on your integration

        Args:
            id_or_slug (int | str): Page ID or slug.
            name (str, optional): Name of page. Defaults to None.
            description (str, optional): A description for this page. Defaults to None.
            amount (int, optional): Default amount you want to accept using this page.
                If none is set, customer is free to provide any amount of their choice.
                The latter scenario is useful for accepting donations. Defaults to None.
            active (bool, optional): Set to false to deactivate page url. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/{id_or_slug}"
        payload = util.generate_payload(locals(), 'id_or_slug')
        if amount == 0:
            payload['amount'] = None
        return self.put(path, payload)

    def check_slug_availability(self, slug: str):
        """Check the availability of a slug for a payment page.

        Args:
            slug (str): URL slug to be confirmed.

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/check_slug_availability/{slug}"
        return self.get(path)

    def add_product(self, page_id: int, product_id: list[int]):
        """Add products to a payment page

        Args:
            page_id (int): ID of the payment page.
            product_id (list[int]): A list of ids of all the product.

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/{page_id}/product"
        payload = {'product': product_id}
        return self.post(path, payload)



from datetime import date, datetime

from ..utilities import decorators, settings, util
from ..utilities.errors import MissingArgumentsError
from ..utilities.request import Request


@decorators.class_type_checker
class Transfer(Request):

    """
    The Transfers API allows you automate sending money on your integration 
    """

    path = '/transfer'

    @staticmethod
    def get_payload(transfer: dict, generate_reference: bool = False) -> dict:
        payload = util.generate_payload(
            transfer, 'source', 'generate_reference')
        payload['recipient'] = util.check_code(
            settings.RECIPIENT, payload['recipient'])

        if 'currency' in payload:
            payload['currency'] = util.check_membership(
                settings.CURRENCIES, payload['currency'], 'currency')

        if not 'reference' in payload and generate_reference:
            payload['reference'] = util.create_ref()

        return payload

    def initiate(self, source: str, amount: int, recipient: str, reason: str = None, currency: str = None, reference: str = None, generate_reference: bool = False):
        """ Status of transfer object returned will be pending if OTP is disabled.
        In the event that an OTP is required, status will read otp.

        Args:
            source (str): Where should we transfer from? Only balance for now
            amount (int): Amount to transfer in kobo if currency is NGN and pesewas if currency is GHS.
            recipient (str): Code for transfer recipient
            reason (str, optional): The reason for the transfer. Defaults to None.
            currency (str, optional): Specify the currency of the transfer. Defaults to NGN. Defaults to None.
            reference (str, optional): If specified, the field should be a unique identifier (in lowercase) for the object.
                Only -,_ and alphanumeric characters allowed. Defaults to None.
            generate_reference (bool, optional): Whether to generate reference for you. Defaults to False.

        Returns:
            JSON: Data fetched from API
        """

        payload = {
            'source': util.check_membership(settings.TRANSFER_SOURCES, source, 'source')
        }
        payload.update(self.get_payload(
            locals(), generate_reference=generate_reference))

        return self.post(self.path, payload)

    def finalize(self, transfer_code: str, otp: str):
        """ Finalize an initiated transfer

        Args:
            transfer_code (str): The transfer code you want to finalize
            otp (str): OTP sent to business phone to verify transfer

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/finalize_transfer'
        payload = {
            'transfer_code': util.check_code(settings.TRANSFER, transfer_code),
            'otp': otp
        }

        return self.post(path, payload)

    def initiate_bulk(self, source: str, *transfers: dict):
        """You need to disable the Transfers OTP requirement to use this endpoint.

        Args:
            source (str): Where should we transfer from? Only balance for now
            transfers (dict): Transfer objects. Each object should contain amount, recipient, and reference

        Raises:
            MissingArgumentsError: raised when a required argument is missing.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/bulk'

        for x in transfers:
            if ('amount', 'recipient') not in x:
                raise MissingArgumentsError(
                    'missing arguments: provide the amount and recipient code')

        payload = {
            'source': util.check_membership(settings.TRANSFER_SOURCES, source, 'source'),
            'transfers': [self.get_payload(transfer, generate_reference=True) for transfer in transfers]
        }

        return self.post(path, payload)

    def list_transfers(self, per_page: int = None, page: int = None, from_date: date | datetime | str = None, to_date: date | datetime | str = None, customer_id: int = None):
        """List the transfers made on your integration.

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page. If not specify we use a default value of 50. Defaults to None.
            page (int, optional): Specify exactly what transfer you want to page. If not specify we use a default value of 1. Defaults to None.
            from_date (date | datetime | str, optional): A timestamp from which to start listing transfer e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            to_date (date | datetime | str, optional): A timestamp at which to stop listing transfer e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            customer_id (int, optional): Filter by customer ID. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)

        if customer_id:
            params.update({'customer': customer_id})

        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch(self, transfer: int | str):
        """Get details of a transfer on your integration.

        Args:
            transfer (int | str): The transfer ID or code you want to fetch

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{util.id_or_code(transfer, data=settings.TRANSFER)}'

        return self.get(path)

    def verify(self, reference: str):
        """Verify the status of a transfer on your integration.

        Args:
            reference (str): Transfer reference

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/verify/{reference}'

        return self.get(path)



from ..utilities import settings, util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class Plan(Request):

    """
    The Plans API allows you create and manage installment payment options on your integration
    """

    path = '/plan'

    def create(self, name: str, amount: int, interval: str, description: str = None, send_invoices: bool = None, send_sms: bool = None, currency: str = None, invoice_limit: int = None):
        """Create a plan on your integration

        Args:
            name (str): Name of plan
            amount (int): Amount should be in kobo if currency is NGN, pesewas,
                if currency is GHS, and cents, if currency is ZAR
            interval (str): Interval in words. Valid intervals are: daily, weekly, monthly,biannually, annually.
            description (str, optional): A description for this plan. Defaults to None.
            send_invoices (bool, optional): Set to false if you don't want invoices to be sent to your customers. Defaults to None.
            send_sms (bool, optional): Set to false if you don't want text messages to be sent to your customers. Defaults to None.
            currency (str, optional): Currency in which amount is set.
                Allowed values are NGN, GHS, ZAR or USD. Defaults to None.
            invoice_limit (int, optional): Number of invoices to raise during subscription to this plan.
                Can be overridden by specifying an invoice_limit while subscribing.. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        payload = util.generate_payload(locals(), 'interval', 'currency')
        payload['interval'] = util.check_membership(
            settings.PLAN_INTERVALS, interval, 'interval')
        if currency:
            payload['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        return self.post(self.path, payload=payload)

    def list_plans(self, per_page: int = None, page: int = None, status: str = None, interval: str = None, amount: int = None):
        """List plans available on your integration.

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.
            page (int, optional): Specify exactly what page you want to retrieve.
                If not specify we use a default value of 1.
            status (str, optional): Filter list by plans with specified status. Defaults to None.
            interval (str, optional): Filter list by plans with specified interval. Defaults to None.
            amount (int, optional): Filter list by plans with specified amount
                ( kobo if currency is NGN, pesewas, if currency is GHS, and cents, if currency is ZAR). Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.check_query_params(per_page=per_page, page=page)

        if interval:
            params['interval'] = util.check_membership(
                settings.PLAN_INTERVALS, interval, 'interval')

        if status:
            params['status'] = util.check_membership(
                settings.PLAN_STATUSES, status, 'status')

        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch(self, plan: int | str):
        """Get details of a plan on your integration.

        Args:
            plan (int | str): The plan ID or code you want to fetch.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{util.id_or_code(plan, data=settings.PLAN)}'

        return self.get(path)

    def update(self, plan: int | str, name: str = None, amount: int = None, interval: str = None, description: str = None, currency: str = None, send_invoices: bool = None, send_sms: bool = None, invoice_limit: int = None):
        """Update a plan details on your integration.

        Args:
            plan (int | str): Code or ID of the plan to update.
            name (str, optional): Name of plan. Defaults to None.
            amount (int, optional): Amount should be in kobo if currency is NGN and pesewas for GHS. Defaults to None.
            interval (str, optional): Interval in words. Valid intervals are hourly, daily, weekly, monthly,biannually, annually. Defaults to None.
            description (str, optional): A description for this plan. Defaults to None.
            send_invoices (bool, optional): Set to false if you don't want invoices to be sent to your customers. Defaults to None.
            send_sms (bool, optional): Set to false if you don't want text messages to be sent to your customers. Defaults to None.
            currency (str, optional): Currency in which amount is set. Any of NGN, GHS, ZAR or USD. Defaults to None.
            invoice_limit (int, optional): Number of invoices to raise during subscription to this plan.
                Can be overridden by specifying an invoice_limit while subscribing.. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{util.id_or_code(plan, data=settings.PLAN)}'
        payload = util.generate_payload(
            locals(), 'plan_id', 'plan_code', 'interval', 'currency')

        if interval:
            payload['interval'] = util.check_membership(
                settings.PLAN_INTERVALS, interval, 'interval')

        if currency:
            payload['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        return self.put(path, payload=payload)



from datetime import date, datetime

from ..utilities import settings, util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class Refund(Request):

    """The Refunds API allows you create and manage transaction refunds
    """

    path = '/refund'

    def create(self, transaction: str | int, amount: int = None, currency: str = None, customer_note: str = None, merchant_note: str = None):
        """Initiate a refund on your integration

        Args:
            transaction (str | int): Transaction reference or id
            amount (int, optional): Amount ( in kobo if currency is NGN, pesewas, if currency is GHS, and cents, if currency is ZAR ) to be refunded to the customer. Amount is optional(defaults to original transaction amount) and cannot be more than the original transaction amount.. Defaults to None.
            currency (str, optional): Three-letter ISO currency. Allowed values are: NGN, GHS, ZAR or USD. Defaults to None.
            customer_note (str, optional): Customer reason. Defaults to None.
            merchant_note (str, optional): Merchant reason. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        payload = util.generate_payload(locals(), 'currency')

        if currency:
            payload['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        return self.post(self.path, payload)

    def list_refunds(self, reference: str = None, currency: str = None, per_page: int = None, page: int = None, from_date: date | datetime | str = None, to_date:  date | datetime | str = None):
        """List refunds available on your integration.

        Args:
            reference (str, optional): Identifier for transaction to be refunded. Defaults to None.
            currency (str, optional): Three-letter ISO currency.
                Allowed values are: NGN, GHS, ZAR or USD. Defaults to None.
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.. Defaults to None.
            page (int, optional): Specify exactly what refund you want to page.
                If not specify we use a default value of 1.. Defaults to None.
            from_date (date | datetime | str, optional): A timestamp from which to start listing refund
                e.g. 2016-09-21. Defaults to None.
            to_date (date | datetime | str, optional): A timestamp at which to stop listing refund
                e.g. 2016-09-21. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)

        if reference:
            params['reference'] = reference

        if currency:
            params['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        if params:
            path = util.handle_query_params(self.path, params)
            return self.get(path)
        return self.get(self.path)

    def fetch(self, reference: str):
        """Get details of a refund on your integration.

        Args:
            reference (str): Identifier for transaction to be refunded

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{reference}'
        return self.get(path)



import http.client
import json
from typing import Sequence
from . import decorators


@decorators.class_type_checker
class Request:

    def __init__(self, secret_key: str) -> None:
        self.headers = {
            'authorization': f'bearer {secret_key}',
            'Content-type': 'application/json',
        }

    @classmethod
    def request(cls, path: str, method: str, headers: dict = None, payload: dict = None):
        connection = http.client.HTTPSConnection('api.paystack.co')
        if payload is not None:
            connection.request(method, path, headers=headers, body=payload)
        else:
            connection.request(method, path, headers=headers)
        return connection.getresponse().read().decode('utf-8')

    def get(self, path: str):
        return self.request(path, 'GET', headers={'authorization': self.headers.pop('authorization')})

    def post(self, path: str, payload: dict | Sequence | set = None):
        if payload:
            return self.request(path, 'POST', headers=self.headers, payload=json.dumps(payload))
        return self.request(path, 'POST', headers=self.headers)

    def put(self, path: str, payload: dict):
        return self.request(path, 'PUT', headers=self.headers, payload=json.dumps(payload))

    def delete(self, path: str, payload: dict = None):
        method = 'DELETE'
        if payload:
            return self.request(path, method, headers=self.headers, payload=payload)
        return self.request(path, method, headers={'authorization': self.headers.pop('authorization')})



from ..utilities import decorators, settings, util
from ..utilities.errors import MissingArgumentsError
from ..utilities.request import Request


@decorators.class_type_checker
class DedicatedVirtualAccounts(Request):

    """
    The Dedicated Virtual Account class enables Nigerian merchants to manage unique payment accounts of their customers.
    """

    path = '/dedicated_account'

    def create(self, customer: int | str, preferred_bank: str, subaccount: str = None, split_code: str = None, first_name: str = None, last_name: str = None, phone: str = None):
        """Create a dedicated virtual account and assign to a customer
            Paystack currently supports Access Bank and Wema Bank.

        Args:
            customer (int | str): Customer ID or code
            preferred_bank (str): The bank slug for preferred bank. To get a list of available banks, use the List Providers endpoint
            subaccount (str, optional): Subaccount code of the account you want to split the transaction with. Defaults to None.
            split_code (str, optional): Split code consisting of the lists of accounts you want to split the transaction with. Defaults to None.
            first_name (str, optional): Customer's first name. Defaults to None.
            last_name (str, optional): Customer's last name. Defaults to None.
            phone (str, optional): Customer's phone number. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        payload = util.generate_payload(
            locals(), 'customer', 'preferred_bank', 'subaccount', 'split_code')
        payload['customer'] = util.id_or_code(customer, data=settings.CUSTOMER)
        payload['preferred_bank'] = util.check_membership(
            settings.VIRTUAL_ACCOUNT_PROVIDERS, preferred_bank, 'preferred_bank')

        if subaccount:
            payload['subaccount'] = util.check_code(
                settings.SUBACCOUNT, subaccount)

        if split_code:
            payload['split_code'] = util.check_code(settings.SPLIT, split_code)

        return self.post(self.path, payload=payload)

    def list_accounts(self, active: bool = None, currency: str = None, provider_slug: str = None, bank_id: int = None, customer_id: str = None):
        """List dedicated virtual accounts available on your integration.

        Args:
            active (bool, optional): Status of the dedicated virtual account. Defaults to None.
            currency (str, optional): The currency of the dedicated virtual account. Only NGN is currently allowed. Defaults to None.
            provider_slug (str, optional): The bank's slug in lowercase, without spaces e.g. wema-bank. Defaults to None.
            bank_id (int, optional): The bank's ID e.g. 035. Defaults to None.
            customer_id (str, optional): The customer's ID. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.generate_payload(locals(), 'currency', 'provider_slug')
        if currency:
            params['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        if provider_slug:
            params['provider_slug'] = util.check_membership(
                settings.VIRTUAL_ACCOUNT_PROVIDERS, provider_slug, 'provider_slug')

        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch(self, dedicated_account_id: int):
        """Get details of a dedicated virtual account on your integration.

        Args:
            dedicated_account_id (int): ID of dedicated virtual account

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{dedicated_account_id}'
        return self.get(path)

    def requery(self, account_number: str, provider_slug: str, date: str = None):
        """Requery Dedicated Virtual Account for new transactions

        Args:
            account_number (str): Virtual account number to requery
            provider_slug (str): The bank's slug in lowercase, without spaces e.g. wema-bank
            date (str, optional): The day the transfer was made in YYYY-MM-DD format. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/requery?account_number={util.check_account_number(account_number)}&provider_slug={util.check_membership(settings.VIRTUAL_ACCOUNT_PROVIDERS, provider_slug, 'provider_slug')}"
        if date:
            path += f'&date={util.handle_date(date)}'
        return self.get(path)

    def deactivate(self, dedicated_account_id: int):
        """Deactivate a dedicated virtual account on your integration.

        Args:
            dedicated_account_id (int): ID of dedicated virtual account

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{dedicated_account_id}'
        return self.delete(path)

    def split_transaction(self, customer: int | str, preferred_bank: str,  subaccount: str | list = None, split_code: str | list = None):
        """Split a dedicated virtual account transaction with one or more accounts

        Args:
            customer (int | str): Customer ID or code.
            preferred_bank (str): The bank slug for preferred bank. To get a list of available banks, use the List Providers endpoint
            subaccount (str | list, optional): Subaccount code of the account you want to split the transaction with. Defaults to None.
            split_code (str | list, optional): Split code consisting of the lists of accounts you want to split the transaction with. Defaults to None.

        Raises:
            MissingArgumentsError: Raised when required arguments are missing.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/split'

        payload = {
            'preferred_bank': util.check_membership(settings.VIRTUAL_ACCOUNT_PROVIDERS, preferred_bank, 'preferred_bank')
        }

        payload['customer'] = util.id_or_code(customer, data=settings.CUSTOMER)

        if not (subaccount or split_code):
            raise MissingArgumentsError(
                'Provide subaccount, split_code, a list of subaccounts or a list of split_codes')

        if subaccount:
            payload['subaccount'] = util.check_code(
                settings.SUBACCOUNT, subaccount)

        if split_code:
            payload['split_code'] = util.check_code(settings.SPLIT, split_code)

        return self.post(path, payload=payload)

    def remove_split(self, account_number: str):
        """If you've previously set up split payment for transactions on a dedicated virtual account,
            you can remove it with this endpoint

        Args:
            account_number (str): Dedicated virtual account number

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/split'
        payload = {
            'account_number': util.check_account_number(account_number)
        }
        return self.delete(path, payload=payload)

    def fetch_bank_providers(self):
        """Get available bank providers for a dedicated virtual account

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/available_providers'
        return self.get(path)



from datetime import date, datetime

from ..utilities import decorators, settings, util
from ..utilities.errors import MissingArgumentsError
from ..utilities.request import Request


@decorators.class_type_checker
class TransferRecipient(Request):

    """
    The Transfer Recipients API allows you create and manage beneficiaries that you send money to
    """

    path = '/transferrecipient'

    @staticmethod
    def get_payload(payload: dict):
        payload = util.generate_payload(payload)
        if not ('recipient_type' or 'type') in payload:
            raise MissingArgumentsError(
                f"provide recipient_type or type based on the type of recipient you want to create, your choices are: {', '.join(settings.RECIPIENT_TYPES)}")

        payload['type'] = util.check_membership(settings.RECIPIENT_TYPES, payload.get(
            'recipient_type') or payload.get('type'), 'type')
        r_type = payload.get('type')

        if r_type == 'authorization':
            if not ('email' and 'authorization_code') in payload:
                raise MissingArgumentsError(
                    "provide email and authorization_code params")

            payload['email'] = util.check_email(payload.pop('email'))
            payload['authorization_code'] = util.check_code(
                settings.AUTHORIZATION, payload.pop('authorization_code'))

            for x in ('bank_code', 'account_number'):
                if x in payload:
                    payload.pop(x)

        elif r_type in ('basa', 'nuban', 'mobile_money'):
            if not ('bank_code' and 'account_number') in payload:
                raise MissingArgumentsError(
                    "provide bank_code and account_number")

            payload['bank_code'] = util.check_bank_code(
                payload.pop('bank_code'))
            payload['account_number'] = util.check_account_number(
                payload.pop('account_number'))

            for x in ('authorization_code', 'email'):
                if x in payload:
                    payload.pop(x)

        currency = payload.get('currency')
        if currency:
            payload['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        return payload

    def create(self, recipient_type: str, name: str, email: str = None, account_number: str = None, bank_code: str = None, description: str = None, currency: str = None, authorization_code: str = None, metadata: dict = None):
        """Creates a new recipient. A duplicate account number will lead to the retrieval of the existing record.

        Args:
            recipient_type (str): Recipient Type. It could be one of: nuban, mobile_money, basa, authorization.
            name (str): A name for the recipient
            email (str, optional): Required if recipient_type is authorization. Defaults to None.
            account_number (str, optional): Required if type is nuban or basa. Defaults to None.
            bank_code (str, optional): Required if type is nuban or basa. You can get the list of Bank Codes by calling the List Banks endpoint. Defaults to None.
            description (str, optional): A description for this recipient. Defaults to None.
            currency (str, optional): Currency for the account receiving the transfer. Defaults to None.
            authorization_code (str, optional): An authorization code from a previous transaction. Defaults to None.
            metadata (dict, optional): Store additional information about your recipient in a structured format, JSON. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        return self.post(self.path, self.get_payload(locals()))

    def bulk_create(self, *recipients: dict[str]):
        """A list of transfer recipient object. Each object should contain type, name, and bank_code.
            Any Create Transfer Recipient param can also be passed.

        Args:
            recipients (dict[str]): Recipient objects. Each object should contain type, name, and bank_code.
                Any Create Transfer Recipient param can also be passed.

        Returns:
            JSON: Data fetched from API
        """
        path = f"{self.path}/bulk"
        payload = {
            'batch': [self.get_payload(recipient) for recipient in recipients]
        }
        return self.post(path, payload)

    def list_recipients(self, per_page: int = None, page: int = None, from_date: date | datetime | str = None, to_date: date | datetime | str = None):
        """List transfer recipients available on your integration

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50. Defaults to None.
            page (int, optional): Specify exactly what page you want to retrieve.
                If not specify we use a default value of 1.. Defaults to None.
            from_date (date | datetime | str, optional): A timestamp from which to start listing transfer recipients
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            to_date (date | datetime | str, optional): A timestamp at which to stop listing transfer recipients
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """
        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)
        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch(self, recipient: int | str):
        """Fetch the details of a transfer recipient.

        Args:
            recipient (int | str): An ID or code for the recipient whose details you want to receive..

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{util.id_or_code(recipient, data=settings.RECIPIENT)}'

        return self.get(path)

    def update(self, recipient: int | str, name: str = None, email: str = None):
        """Update an existing recipient. An duplicate account number will lead to the retrieval of the existing record.
            
        Args:
            recipient (int | str): Transfer Recipient's ID or code
            name (str, optional): A name for the recipient. Defaults to None.
            email (str, optional): Email address of the recipient. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """
        path = f'{self.path}/{util.id_or_code(recipient, data=settings.RECIPIENT)}'

        payload = {}

        if name:
            payload['name'] = name

        if email:
            payload['email'] = util.check_email(email)

        return self.put(path, payload)

    def delete(self, recipient: int | str):
        """Deletes a transfer recipient (sets the transfer recipient to inactive).
        
        Args:
            recipient (int | str): An ID or code for the recipient who you want to delete.
        
        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{util.id_or_code(recipient, data=settings.RECIPIENT)}'

        return super().delete(path)



from ..utilities import decorators, settings, util
from ..utilities.request import Request


@decorators.class_type_checker
class TransferControl(Request):

    """
    The Transfers Control API allows you manage settings of your transfers
    """

    balance = '/balance'
    transfer = '/transfer'

    def check_balance(self):
        """Fetch the available balance on your integration

        Returns:
            JSON: Data fetched from API
        """

        return self.get(self.balance)

    def fetch_balance_ledger(self):
        """Fetch all pay-ins and pay-outs that occured on your integration

        Returns:
            JSON: Data fetched from API
        """
        path = f'{self.balance}/ledger'

        return self.get(path)

    def resend_otp(self, transfer_code: str, reason: str):
        """Generates a new OTP and sends to customer in the event they are having trouble receiving one.
            This feature is only available to businesses in Nigeria and Ghana.

        Args:
            transfer_code (str): Transfer code
            reason (str): Either resend_otp or transfer

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.transfer}/resend_otp'
        payload = {
            'transfer_code': util.check_code(settings.TRANSFER, transfer_code),
            'reason': util.check_membership(settings.TRANSFER_CONTROL_REASONS, reason, 'reason')
        }

        return self.post(path, payload)

    def disable_otp(self):
        """This is used in the event that you want to be able to complete transfers programmatically without use of OTPs.
            No arguments required. You will get an OTP to complete the request.
            This feature is only available to businesses in Nigeria and Ghana.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.transfer}/disable_otp'

        return self.post(path, {})

    def finalize_disable_otp(self, otp: str):
        """Finalize the request to disable OTP on your transfers.
            This feature is only available to businesses in Nigeria and Ghana.

        Args:
            otp (str): OTP sent to business phone to verify disabling OTP requirement

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.transfer}/disable_otp_finalize'

        return self.post(path, {'otp': otp})

    def enable_otp(self):
        """In the event that a customer wants to stop being able to complete transfers programmatically,
        this endpoint helps turn OTP requirement back on. No arguments required.
        This feature is only available to businesses in Nigeria and Ghana.

        Returns:
           JSON: Data fetched from API
        """
        
        path = f'{self.transfer}/enable_otp'

        return self.post(path)



from datetime import date, datetime

from ..utilities import util, decorators, settings
from ..utilities.request import Request


@decorators.class_type_checker
class Settlements(Request):

    """
    The Settlements API allows you gain insights into payouts made by Paystack to your bank account
    """

    path = '/settlement'

    def fetch(self, per_page: int = None, page: int = None, from_date: date | datetime | str = None, to_date: date | datetime | str = None, subaccount: str = None):
        """Fetch settlements made to your settlement accounts.

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.
            page (int, optional): Specify exactly what page you want to retrieve.
                If not specify we use a default value of 1.
            from_date (date | datetime | str, optional): A timestamp from which to start listing settlements
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            to_date (date | datetime | str, optional): A timestamp at which to stop listing settlements
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            subaccount (str, optional): Provide a subaccount ID to export only settlements for that subaccount.
                Set to none to export only transactions for the account. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)
        params.update({'subaccount': util.check_code(
            settings.SUBACCOUNT, subaccount)})

        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch_transactions(self, settlement_id: int, per_page: int = None, page: int = None, from_date: date | datetime | str = None, to_date: date | datetime | str = None):
        """Get the transactions that make up a particular settlement

        Args:
            settlement_id (int): The settlement ID in which you want to fetch its transactions.
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.. Defaults to None.
            page (int, optional): Specify exactly what page you want to retrieve.
                If not specify we use a default value of 1.. Defaults to None.
            from_date (date | datetime | str, optional): A timestamp from which to start listing settlement transactions
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            to_date (date | datetime | str, optional): A timestamp at which to stop listing settlement transactions
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/{settlement_id}/transactions"
        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)
        if params:
            return self.get(util.handle_query_params(path, params))
        return self.get(path)



from datetime import date, datetime

from ..utilities import settings, util, decorators
from ..utilities.errors import MissingArgumentsError, UnwantedArgumentsError
from ..utilities.request import Request


@decorators.class_type_checker
class Charge(Request):

    """
    The Charge API allows you to configure payment channel of your choice when initiating a payment.
    """

    path = '/charge'

    def create(self, email: str, amount: int, bank_code: str = None, account_number: str = None, currency: str = None, authorization_code: str = None, reference: str = None, pin: str = None, ussd: int = None, phone: str = None, provider: str = None, metadata: dict = None, device_id: str = None, generate_reference: bool = False):
        """The Charge API allows you to configure payment channel of your choice when initiating a payment.

        Args:
            email (str): Customer's email address
            amount (int): Amount should be in kobo if currency is NGN, pesewas, if currency is GHS, and cents, if currency is ZAR
            bank_code (str, optional): Bank code to use. Provide with account_number. Don't send if you sending authorization_code. Defaults to None.
            account_number (str, optional): Account number to charge. Provide with bank_code. Don't send if you sending authorization_code. Defaults to None.
            currency (str, optional): The currency you want to charge in. Defaults to None.
            authorization_code (str, optional): An authorization code to charge (don't send if charging a bank account). Defaults to None.
            reference (str, optional): Unique transaction reference. Only -, .`, = and alphanumeric characters allowed.. Defaults to None.
            pin (str, optional): 4-digit PIN (send with a non-reusable authorization code). Defaults to None.
            ussd (int, optional): USSD type to charge (don't send if charging an authorization code, bank or card). Defaults to None.
            phone (str, optional): 11 digit str phone number to use with mobile money. Provide with provider. Defaults to None.
            provider (str, optional): mobile money provider, your choices are 'tgo' for Airtel/Tigo, 'mtn' for MTN, 'vod' for vodafone. Defaults to None.
            metadata (dict, optional): A JSON object. Defaults to None.
            device_id (str, optional): This is the unique identifier of the device a user uses in making payment. Only -, .`, = and alphanumeric characters allowed.. Defaults to None.
            generate_reference (bool, optional): Whether to generate reference for you. Defaults to False.

        Raises:
            MissingArgumentsError: raised when a wanted argument is missing

            UnwantedArgumentsError: raised when an unwanted argument is provided

        Returns:
            json: Data fetched from API
        """

        payload = {
            'email': util.check_email(email),
            'amount': amount
        }

        methods = {bank_code, account_number,
                   authorization_code, pin, ussd, phone, provider}

        if metadata:
            payload['metadata'] = metadata

        if currency:
            payload['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        if reference:
            payload['reference'] = reference

        if not reference and generate_reference:
            payload['reference'] = util.create_ref()

        if device_id:
            payload['device_id'] = device_id

        if bank_code or account_number:
            if not (bank_code and account_number):
                raise MissingArgumentsError(
                    'provide both the bank_code and account_number')

            unwanted = methods.difference({bank_code, account_number})
            if any(unwanted):
                raise UnwantedArgumentsError(
                    f"don't provide these: {', '.join(unwanted)} - if you're going use bank_code and account_number")

            payload.update({'bank': {'code': util.check_bank_code(
                bank_code), 'account_number': util.check_account_number(account_number)}})
            return self.post(self.path, payload)

        if ussd:
            unwanted = methods.difference({ussd})
            if any(unwanted):
                raise UnwantedArgumentsError(
                    f"don't provide these: {', '.join(unwanted)}- if you're going to use ussd")

            payload.update({'ussd': util.check_membership(
                settings.USSD_CODES, ussd, 'ussd code')})
            return self.post(self.path, payload)

        if pin:
            if not (pin and authorization_code):
                raise MissingArgumentsError(
                    'provide pin with a non-reuseable authorization code')

            unwanted = methods.difference({pin, authorization_code})
            if any(unwanted):
                raise UnwantedArgumentsError(
                    f"don't provide these: {', '.join(unwanted)} - if you're going to use pin and authorization_code")
            payload.update(
                {'pin': pin, 'authorization_code': authorization_code})
            return self.post(self.path, payload)

        if authorization_code:
            unwanted = methods.difference({authorization_code})
            if any(unwanted):
                raise UnwantedArgumentsError(
                    f"don't provide these: {', '.join(unwanted)} - if you're going to use authorization_code")
            payload.update({'authorization_code': authorization_code})
            return self.post(self.path, payload)

        if phone or provider:
            if not (phone and provider):
                raise MissingArgumentsError('provide phone and provider')

            unwanted = methods.difference({phone, provider})
            if any(unwanted):
                raise UnwantedArgumentsError(
                    f"don't provide these: {', '.join(unwanted)} - if you're going to use phone and provider for mobile money option")
            payload.update({'mobile_money': {'phone': phone, 'provider': util.check_membership(
                settings.MOBILE_PAYMENT_PROVIDERS, provider, 'provider')}})
            return self.post(self.path, payload)

    def submit_pin(self, pin: str, reference: str):
        """Submit PIN to continue a charge

        Args:
            pin (str): Pin provided by the user
            reference (str): reference of the transaction that requested the pin

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/submit_pin'
        payload = util.generate_payload(locals())
        return self.post(path, payload)

    def submit_otp(self, otp: str, reference: str):
        """Submit OTP to complete a charge

        Args:
            otp (str): OTP submitted by the user
            reference (str): Reference for the ongoing transation

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/submit_otp"
        payload = util.generate_payload(locals)
        return self.post(path, payload)

    def submit_phone(self, phone: str, reference: str):
        """Submit Phone when requested

        Args:
            phone (str): Phone number submitted by the user
            reference (str): Reference of the ongoing transation

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/submit_phone"
        payload = util.generate_payload(locals())
        return self.post(path, payload)

    def submit_birthday(self, birthday: date | datetime | str, reference: str):
        """Submit Phone when requested

        Args:
            birthday ( date | datetime | str ): birthday submitted by the user
            reference (str): Reference of the ongoing transaction

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/submit_birthday"
        payload = util.generate_payload(locals())
        return self.post(path, payload)

    def submit_address(self, reference: str, address: str, city: str, state: str, zipcode: str):
        """Submit address to continue charge

        Args:
            reference (str): Reference for the ongoing transaction
            address (str): Address submitted by the user
            city (str): City submitted by the user
            state (str): State submitted by the user
            zipcode (str): Zipcode submitted by the user

        Returns:
            JSON: Data fetched from paystack API
        """
        path = f"{self.path}/submit_address"
        payload = util.generate_payload(locals())
        return self.post(path, payload)

    def check_pending_charge(self, reference: str):
        """When you get "pending" as a charge status or if there was an exception when calling any of the /charge endpoints,
        wait 10 seconds or more, then make a check to see if its status has changed.
        Don't call too early as you may get a lot more pending than you should.

        Args:
            reference (str): Reference to check

        Returns:
            JSON: Data fetched from API
        """
        path = f"{self.path}/{reference}"
        return self.get(path)



from collections import namedtuple

Code = namedtuple('Code', ['prefix', 'name'])
SUBACCOUNT = Code('ACCT', 'subaccount')
PLAN = Code('PLN', 'plan')
SUBSCRIPTION = Code('SUB', 'subscription')
AUTHORIZATION = Code('AUTH', 'authorization')
SPLIT = Code('SPL', 'split')
PRODUCT = Code('PROD', 'product')
INVOICE = Code('PRQ', 'invoice')
RECIPIENT = Code('RCP', 'recipient')
TRANSFER = Code('TRF', 'transfer')
BULK_CHARGE_BATCH = Code('BCH', 'bulk_charge_batch')
CUSTOMER = Code('CUS', 'customer')

DOCUMENT_TYPES = ('identityNumber', 'passportNumber',
                  'businessRegistrationNumber')

ACCOUNT_TYPES = ('personal', 'business')

PAYMENT_CHANNELS = ('card', 'bank', 'ussd', 'qr',
                    'mobile_money', 'bank_transfer')

CURRENCIES = ('GHS', 'NGN', 'USD', 'ZAR')

TRANSACTION_STATUS = ('failed', 'success', 'abandoned')

SPLIT_TYPES = ('percentage', 'flat')

BEARER_TYPES = ('account', 'subaccount', 'all-proportional', 'all')

IDENTIFICATION_TYPES = ('bank_account', 'bvn')

RISK_ACTIONS = ('default', 'allow', 'deny')

VIRTUAL_ACCOUNT_PROVIDERS = ('access-bank', 'wema-bank')

SETTLEMENT_SCHEDULES = ('auto', 'weekly', 'monthly')

PLAN_INTERVALS = ('daily', 'weekly', 'monthly', 'biannually', 'annually')

PLAN_STATUSES = ('active', 'complete', 'non-renewing',
                 'attention', 'cancelled')

RECIPIENT_TYPES = ('nuban', 'mobile_money', 'basa', 'authorization')

TRANSFER_SOURCES = ('balance',)

TRANSFER_CONTROL_REASONS = ('resend_otp', 'transfer')

BULK_CHARGE_STATUSES = ('active', 'paused', 'complete')

USSD_CODES = (737, 919, 822, 966)

MOBILE_PAYMENT_PROVIDERS = ("mtn", "vod", "tgo")

DISPUTE_STATUSES = ('awaiting-merchant-feedback',
                    'awaiting-bank-feedback', 'pending', 'resolved')

DISPUTE_RESOLUTION = ('merchant-accepted', 'declined')

BANK_GATEWAYS = ('emandate', 'digitalbankmandate')



from ..utilities import util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class ApplePay(Request):

    """
    The Apple Pay API allows you register your application's top-level domain or subdomain
    """

    path = '/apple-pay/domain'

    def register(self, domain: str):

        """Register a top-level domain or subdomain for your Apple Pay integration.

        Args:
            domain [str]: The domain name to be registered.

        Returns:
            json: Data fetched from paystack API
        """

        payload = {
            'domainName': util.check_domain(domain)
        }
        return self.post(self.path, payload=payload)

    def list_domains(self):

        """Lists all registered domains on your integration.
        Returns an empty array if no domains have been added.

        Returns:
            json: Data fetched from paystack API
        """

        return self.get(self.path)

    def unregister(self, domain: str):

        """Unregister a top-level domain or subdomain previously used for your Apple Pay integration.

        Args:
            domain [str]: The domain name to be unregistered

        Returns:
            json: Data fetched from paystack API
        """

        payload = {
            'domainName': util.check_domain(domain)
        }
        return self.delete(self.path, payload=payload)



import datetime

from ..utilities import settings, util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class Subscription(Request):

    """
    The Subscriptions API allows you create and manage recurring payment on your integration
    """

    path = "/subscription"

    def create(self, email_or_customer_code: str, plan_code: str, authorization_code: str = None, start_date: datetime.datetime | datetime.date | str = None):
        """Create a subscription on your integration

        Args:
            email_or_customer_code (str): Customer's email address or customer code
            plan_code (str): Plan code
            authorization_code (str, optional): If customer has multiple authorizations,
                you can set the desired authorization you wish to use for this subscription here.
                If this is not supplied, the customer's most recent authorization would be used. Defaults to None.
            start_date (datetime.datetime | datetime.date | str, optional): Set the date for the first debit. 
                (ISO 8601 format) e.g. 2017-05-16T00:30:13+01:00. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        payload = {
            'customer': util.check_email_or_customer(email_or_customer_code),
            'plan': util.check_code(settings.PLAN, plan_code)
        }

        if authorization_code:
            payload['authorization'] = util.check_code(
                settings.AUTHORIZATION, authorization_code)

        if start_date:
            payload['start_date'] = util.handle_date(start_date)

        return self.post(self.path, payload=payload)

    def list_subscriptions(self, per_page: int = None, page: int = None, customer: int = None, plan: int = None):
        """List subscriptions available on your integration.

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.. Defaults to None.
            page (int, optional): Specify exactly what page you want to retrieve.
                If not specify we use a default value of 1.. Defaults to None.
            customer (int, optional): Filter by Customer ID. Defaults to None.
            plan (int, optional): Filter by Plan ID. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """
        params = util.generate_payload(locals())
        params.update(util.check_query_params(per_page=per_page, page=page))

        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch(self, subscription: int | str):
        """Get details of a subscription on your integration.

        Args:
            subscription (int | str): The subscription ID or code you want to fetch.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{util.id_or_code(subscription, data=settings.SUBSCRIPTION)}'
        
        return self.get(path)

    def enable(self, subscription_code: str, email_token: str):
        """Enable a subscription on your integration

        Args:
            subscription_code (str): Subscription code
            email_token (str): Email token

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/enable"
        payload = {
            'code': util.check_code(settings.SUBSCRIPTION, subscription_code),
            'token': email_token
        }

        return self.post(path, payload)

    def disable(self, subscription_code: str, email_token: str):
        """Disable a subscription on your integration

        Args:
            subscription_code (str): Subscription code
            email_token (str): Email token


        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/disable"
        payload = {
            'code': util.check_code(settings.SUBSCRIPTION, subscription_code),
            'token': email_token
        }
        return self.post(path, payload)

    def generate_update_link(self, subscription_code: str):
        """Generate a link for updating the card on a subscription

        Args:
            subscription_code (str): Subscription code

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/{util.check_code(settings.SUBSCRIPTION, subscription_code)}/manage/link"

        return self.get(path)

    def send_update_link(self, subscription_code: str):
        """ Email a customer a link for updating the card on their subscription

        Args:
            subscription_code (str): Subscription code

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/{util.check_code(settings.SUBSCRIPTION, subscription_code)}/manage/link"

        return self.get(path)



from ..utilities.request import Request
from ..utilities import decorators


@decorators.class_type_checker
class ControlPanel(Request):

    """
    The Control Panel API allows you manage some settings on your integration
    """

    path = 'integration/payment_session_timeout'

    def fetch_payment_session_timeout(self):
        """Fetch the payment session timeout on your integration

        Returns:
            JSON: Data fetched from API
        """
        return self.get(self.path)

    def update_payment_session_timeout(self, time_out: int):
        """Update the payment session timeout on your integration

        Args:
            time_out (int): Time before stopping session (in seconds).
                Set to 0 to cancel session timeouts

        Returns:
            JSON: Data fetched from API
        """
        return self.put(self.path, {'timeout': time_out})



from ..utilities import settings, util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class Verification(Request):

    """
    The Verification class allows you perform KYC ( Know Your Customer ) processes.
    This feature is only available to businesses in Nigeria.
    """

    def resolve_acct_number(self, account_number: str, bank_code: str):
        """Confirm an account belongs to the right customer

        Args:
            account_number (str): Account Number
            bank_code (str): You can get the list of bank codes by calling the List Bank endpoint

        Returns:
            JSON: Data fetched from API
        """

        path = f'/bank/resolve?account_number={util.check_account_number(account_number)}&bank_code={util.check_bank_code(bank_code)}'
        return self.get(path)

    def validate_account(self, account_name: str, account_number: str, account_type: str, bank_code: str, country_code: str, document_type: str, document_number: str):
        """Confirm the authenticity of a customer's account number before sending money

        Args:
            account_name (str): Customer's first and last name registered with their bank
            account_number (str): Customer's account number
            account_type (str): This can take one of: [ personal, business ]
            bank_code (str): The bank code of the customer's bank. You can fetch the bank codes by using our List Bank endpoint
            country_code (str): The two digit ISO code of the customer's bank e.g GH for Ghana
            document_type (str): Customer's mode of identity. This could be one of: [ identityNumber, passportNumber, businessRegistrationNumber ]
            document_number (str): Customer's mode of identity number.

        Returns:
            JSON: Data fetched from API
        """

        path = '/bank/validate'
        payload = {
            'account_name': account_name,
            'account_number': util.check_account_number(account_number),
            'account_type': util.check_membership(settings.ACCOUNT_TYPES, account_type, 'account_type'),
            'bank_code': util.check_bank_code(bank_code),
            'country_code': util.check_country(country_code),
            'document_type': util.check_membership(settings.DOCUMENT_TYPES, document_type, 'document_type'),
            'document_number': document_number
        }
        return self.post(path, payload=payload)

    def resolve_card_bin(self, card_bin: str):
        """Get more information about a customer's card

        Args:
            card_bin (str): First 6 characters of card

        Raises:
            ValueError: raised when card_bin is less or greater than 6
            or is not made up of digits.

        Returns:
            JSON: Data fetched from API
        """

        path = '/decision/bin/{}'
        if len(card_bin) != 6 or not card_bin.isdigit():
            raise ValueError('value for card_bin must be 6 digits')
        return self.get(path.format(card_bin))



from inspect import signature, _empty
from functools import wraps
from types import GenericAlias


def func_type_checker(func):

    @wraps(func)
    def wrapper(*args, **kwargs):
        sig = signature(func)
        bound = sig.bind(*args, **kwargs)
        for key, value in bound.arguments.items():
            ann = sig.parameters[key].annotation

            if ann == _empty or isinstance(ann, GenericAlias):
                continue

            if not isinstance(value, ann):
                class_name = bound.arguments.get('self')
                func_name = func.__name__
                if class_name:
                    func_name = f"{class_name.__class__.__name__}.{func_name}"

                raise TypeError(
                    f"Expected type {ann} for {key} in function {func_name}, got {type(value)} instead")
        return func(*args, **kwargs)
    return wrapper


def class_type_checker(_cls):

    class C(_cls):

        def __new__(cls, *args, **kwargs):
            for key, value in _cls.__dict__.items():
                if callable(value):
                    setattr(_cls, key, func_type_checker(value))
            return super().__new__(cls)
    C.__name__ = _cls.__name__
    C.__doc__ = _cls.__doc__
    return C


@func_type_checker
def test(a: list[dict[str]]):
    return a



import datetime

from ..utilities import settings, util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class Customer(Request):

    """
    The Customers class allows you create and manage customers on your integration.
    """

    path = '/customer'

    def create(self, email: str, first_name: str = None, last_name: str = None, phone: str = None, metadata: dict = None):
        """Create a customer on your integration. The first_name, last_name and phone are optional parameters.
            However, when creating a customer that would be assigned a Dedicated Virtual Account and your business catgeory falls under Betting,
            Financial services, and General Service, then these parameters become compulsory.

        Args:
            email (str): Customer's email address.
            first_name (str, optional): Customer's first name. Defaults to None.
            last_name (str, optional): Customer's last name. Defaults to None.
            phone (str, optional): Customer's phone number. Defaults to None.
            metadata (dict, optional): A set of key/value pairs that you can attach to the customer.
                It can be used to store additional information in a structured format. Defaults to None.

        Returns:
            JSON: Data fetched from API.
        """

        payload = util.generate_payload(locals(), 'email')
        payload['email'] = util.check_email(email)
        return self.post(self.path, payload)

    def list_customers(self, per_page: int = None, page: int = None, from_date: datetime.datetime | datetime.date | str = None, to_date: datetime.datetime | datetime.date | str = None):
        """List customers available on your integration.

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50. Defaults to None.
            page (int, optional): Specify exactly what page you want to retrieve.
                If not specify we use a default value of 1. Defaults to None.
            from_date (datetime.datetime | datetime.date | str, optional): A timestamp from which to start listing customers
                e.g. 2016-09-24T00:00:05.000Z, 2016. Defaults to None.
            to_date (datetime.datetime | datetime.date | str, optional): A timestamp at which to stop listing customers
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)
        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch(self, email_or_customer_code: str):
        """Get details of a customer on your integration.

        Args:
            email_or_customer_code (str): An email or customer code for the customer you want to fetch

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{util.check_email_or_customer(email_or_customer_code)}'
        return self.get(path)

    def update(self, customer_code: str, first_name: str = None, last_name: str = None, phone: str = None, metadata: dict = None):
        """Update a customer's details on your integration

        Args:
            customer_code (str): Code for the customer you want to update.
            first_name (str, optional): Customer's first name. Defaults to None.
            last_name (str, optional): Customer's last name. Defaults to None.
            phone (str, optional): Customer's phone number. Defaults to None.
            metadata (dict, optional): A set of key/value pairs that you can attach to the customer.
                It can be used to store additional information in a structured format. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{util.check_code(settings.CUSTOMER, customer_code)}'
        payload = util.generate_payload(locals(), 'customer_code')
        return self.put(path, payload=payload)

    def validate(self, customer_code: str, first_name: str, last_name: str, identification_type: str, country: str, bvn: str, bank_code: str = None, account_number: str = None, middle_name: str = None):
        """Validate a customer's identity

        Args:
            customer_code (str): Code for the customer you want to validate
            first_name (str): Customer's first name
            last_name (str): Customer's last name
            identification_type (str): Predefined types of identification. Valid values: bvn, bank_account
            country (str): 2 letter country code of identification issuer
            bvn (str): Customer's Bank Verification Number
            bank_code (str, optional): You can get the list of Bank Codes by calling the List Banks endpoint. (required if type is bank_account). Defaults to None.
            account_number (str, optional): Customer's bank account number. (required if type is bank_account). Defaults to None.
            middle_name (str, optional): Customer's middle name. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{util.check_code(settings.CUSTOMER, customer_code)}/identification'
        payload = {
            'first_name': first_name,
            'last_name': last_name,
            'type': util.check_membership(settings.IDENTIFICATION_TYPES, identification_type, 'identification_type'),
            'country': util.check_country(country),
            'bvn': util.check_bvn(bvn)
        }

        if payload['type'] == 'bank_account':
            assert bank_code and account_number, "You'll need the customers bank_code and account_number for this type of verification"
            payload.update(
                {'bank_code': bank_code, 'account_number': account_number})

        if middle_name is not None:
            payload['middle_name'] = middle_name

        return self.post(path, payload=payload)

    def whitelist_blacklist(self, email_or_customer_code: str, risk_action: str):
        """Whitelist or blacklist a customer on your integration

        Args:
            email_or_customer_code (str): Customer's code or email address
            risk_action (str): One of the possible risk actions [ default, allow, deny ].
                allow to whitelist. deny to blacklist. Customers start with a default risk action.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/set_risk_action'

        payload = {
            'risk_action': util.check_membership(settings.RISK_ACTIONS, risk_action, 'risk_action'),
            'customer': util.check_email_or_customer(email_or_customer_code)
        }

        return self.post(path, payload=payload)

    def deactivate_authorization(self, authorization_code: str):
        """Deactivate an authorization when the card needs to be forgotten

        Args:
            authorization_code (str): Authorization code for a card that needs to be forgotten.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/deactivate_authorization'

        payload = {'authorization_code': util.check_code(
            settings.AUTHORIZATION, authorization_code)}

        return self.post(path, payload=payload)



import datetime
from typing import Iterable

from ..utilities import settings, util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class TransactionSplit(Request):
    """
    Create, list, retrieve, update split transaction configuration with one or more SubAccounts (You should have subaccounts on your integration to use this)  
    """

    path = '/split'

    def create(self, name: str, split_type: str, currency: str, subaccounts: Iterable[tuple[str, int]], bearer_type: str, bearer_subaccount: str):
        """Create a split payment on your integration

        Args:
            name (str): Name of the transaction split
            split_type (str): The type of transaction split you want to create. 
                You can use one of the following: percentage | flat
            currency (str): Any of NGN, GHS, ZAR, or USD
            subaccounts (Iterable[tuple[str, int]]): A list of object containing subaccount code and
                number of shares: [('ACT_xxxxxxxxxx', xxx)]
            bearer_type (str): Any of subaccount | account | all-proportional | all
            bearer_subaccount (str): Subaccount code

        Returns:
            JSON: Data fetched from API
        """

        payload = {
            'name': name,
            'type': util.check_membership(settings.SPLIT_TYPES, split_type, 'split_type'),
            'currency': util.check_membership(settings.CURRENCIES, currency, 'currency'),
            'bearer_type': util.check_membership(settings.BEARER_TYPES, bearer_type, 'bearer_type'),
            'subaccounts': [{'subaccount': util.check_code(settings.SUBACCOUNT, subaccount), 'share': share} for subaccount, share in subaccounts],
        }

        if bearer_type == 'subaccount':
            payload['bearer_subaccount'] = util.check_code(
                settings.SUBACCOUNT, bearer_subaccount)

        return self.post(self.path, payload=payload)

    def list_search(self, name: str = None, active: bool = None, sort_by: str = None, per_page: int = None, page: int = None, from_date: datetime.datetime | datetime.date | str = None, to_date: datetime.datetime | datetime.date | str = None):
        """List/search for the transaction splits available on your integration.

        Args:
            name (str, optional): The name of the split. Defaults to None.
            active (bool, optional): Any of true or false. Defaults to None.
            sort_by (str, optional): Sort by name, defaults to createdAt date. Defaults to None.
            per_page (int, optional): Number of splits per page. If not specify we use a default value of 50.
            page (int, optional): Page number to view. If not specify we use a default value of 1.
            from_date (datetime.datetime | datetime.date | str, optional): A timestamp from which to start listing splits
                e.g. 2019-09-24T00:00:05.000Z, 2019-09-21. Defaults to None.
            to_date (datetime.datetime | datetime.date | str, optional): A timestamp at which to stop listing splits
                e.g. 2019-09-24T00:00:05.000Z, 2019-09-21. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.generate_payload(
            locals(), 'per_page', 'page', 'from_date', 'to_date')
        params.update(util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date))
        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch(self, split_id: int):
        """Get details of a split on your integration.

        Args:
            split_id (int): The id of the split

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{split_id}'
        return self.get(path)

    def update(self, split_id: int, name: str = None, active: bool = None, bearer_type: str = None, bearer_subaccount: str = None):
        """Update a transaction split details on your integration

        Args:
            split_id (int): Split ID
            name (str, optional): Name of the transaction split. Defaults to None.
            active (bool, optional): True or False. Defaults to None.
            bearer_type (str, optional): Any of the following values:
                subaccount | account | all-proportional | all. Defaults to None.
            bearer_subaccount (str, optional): Subaccount code of a subaccount in the split group.
                This should be specified only if the bearer_type is subaccount. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{split_id}'
        payload = {key: value for key, value in locals(
        ).items() if key not in ('split_id', 'path') and value is not None}

        if bearer_type:
            payload['bearer_type'] = util.check_membership(
                settings.BEARER_TYPES, bearer_type, 'bearer_type')

        if 'bearer_type' in payload and payload['bearer_type'] != 'subaccount':
            payload.pop('bearer_subaccount')

        return self.put(path, payload=payload)

    def add_update_subaccount(self, split_id: int, subaccount: str, share: int):
        """Add a Subaccount to a Transaction Split, or update the share of an existing Subaccount in a Transaction Split

        Args:
            split_id (int): Split Id
            subaccount (str): This is the sub account code
            share (int): This is the transaction share for the subaccount

        Returns:
            JSON: Data fetched from API   
        """

        path = f'{self.path}/{split_id}/subaccount/add'
        payload = {
            'subaccount': util.check_code(settings.SUBACCOUNT, subaccount),
            'share': share
        }
        return self.post(path, payload=payload)

    def remove_subaccount(self, split_id: int, subaccount: str):
        """Remove a subaccount from a transaction split

        Args:
            split_id (int): Split Id
            subaccount (str): This is the sub account code

        Returns:
            JSON: Data fetched from API        
        """

        path = f'{self.path}/{split_id}/subaccount/remove'
        payload = {
            'subaccount': util.check_code(settings.SUBACCOUNT, subaccount)
        }
        return self.post(path, payload=payload)



import datetime
import re
import uuid
from typing import Sequence

from . import settings
from .errors import MissingArgumentsError


def check_bvn(bvn: str) -> str:
    if not bvn.isdigit() or len(bvn) != 11:
        raise ValueError("bvn must contain only digits and be 11 digits long")
    return bvn


def create_ref() -> str:
    return str(uuid.uuid4()).replace('-', '')


def check_channels(channel) -> list:
    channels = settings.PAYMENT_CHANNELS
    message = f"Invalid payment channel, choices are: {', '.join(channels)}"
    if isinstance(channel, list):
        if not set(channel).issubset(channels):
            raise ValueError(message)
    else:
        if not channel in channels:
            raise ValueError(message)
    return [channel]


def handle_date(date: datetime.date | datetime.datetime | str) -> str:
    return date.strftime('%Y-%m-%d') if isinstance(date, (datetime.date, datetime.datetime)) else datetime.datetime.strptime(date, '%Y-%m-%d').strftime('%Y-%m-%d')


def check_email(email: str) -> bool:
    if not re.fullmatch(r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-z|a-z]{2,}\b', email):
        raise ValueError('provide a valid email!')
    return email


def check_country(country_code: str) -> str:
    if len(country_code) != 2:
        raise ValueError(
            "Invalid country_code, value must be a 2 character string")
    return country_code


def check_account_number(account_number: str) -> str:
    if len(account_number) != 10:
        raise ValueError(
            "Invalid account number, value must be a 10 character string")
    return account_number


def check_email_or_customer(email_or_customer_code: str) -> str:
    try:
        customer = check_email(email_or_customer_code)
    except ValueError:
        try:
            customer = check_code(email_or_customer_code, 'CUS')
        except ValueError as error:
            raise ValueError(
                'Invalid value for email_or_customer_code, provide an email or customer_code') from error
    return customer


def check_bank_code(bank_code: str) -> str:
    if not bank_code.isdigit():
        raise ValueError("bank_code should be a string of digits")
    return bank_code


def check_domain(domain: str) -> str:
    pattern = re.compile(
        r'^(([a-zA-Z]{1})|([a-zA-Z]{1}[a-zA-Z]{1})|'
        r'([a-zA-Z]{1}[0-9]{1})|([0-9]{1}[a-zA-Z]{1})|'
        r'([a-zA-Z0-9][-_.a-zA-Z0-9]{0,61}[a-zA-Z0-9]))\.'
        r'([a-zA-Z]{2,13}|[a-zA-Z0-9-]{2,30}.[a-zA-Z]{2,3})$'
    )
    if not re.match(pattern, domain):
        raise ValueError('Invalid domain')
    return domain


def check_query_params(per_page: int = None, page: int = None, from_date: datetime.date | datetime.datetime | str = None, to_date: datetime.date | datetime.datetime | str = None) -> dict:
    params = {}
    if per_page:
        params['perPage'] = per_page

    if page:
        params['page'] = page

    if from_date:
        params['from'] = handle_date(from_date)

    if to_date:
        params['to'] = handle_date(to_date)

    return params


def handle_query_params(path: str, params: dict) -> str:
    path += '?'
    for key, value in params.items():
        path += f"{key}={value}&"
    return path.rstrip('&')


def check_code(data: tuple[str, str], code: Sequence[str] | str) -> str | list:
    if isinstance(code, (str)):
        code = tuple(code)

    for x in code:
        pre = x.split('_')[0]
        if pre != data.prefix:
            raise ValueError("Invalid {0}: {0} must start with {1} not {2}".format(
                data.name, data.prefix, pre))
    if len(code) == 1:
        return code[0]
    return code


def check_membership(group: Sequence, member: str, name: str) -> str:
    if member not in group:
        raise ValueError(
            f"Invalid value for {name}, your choices are: {', '.join(group)}")
    return member


def generate_payload(payload: dict, *unwanted) -> dict:
    rem = ['self', 'path']
    rem.extend(unwanted)
    return {key: value for key, value in payload.items() if value is not None and key not in rem}


def id_or_code(value: str | int, data: tuple = None):
    if isinstance(value, int):
        return value
    return check_code(data, value)



from .routes import (apple_pay, bulk_charges, charge, control_panel, customers,
                     disputes, invoices, miscellaneous, payment_pages, plans,
                     products, refund, settlement, subaccount, subscription,
                     transactions, transactions_split, transfer_recipient,
                     transfers, transfers_control, verification,
                     virtual_accounts)
from .utilities import decorators


class Paystack:

    """
    General class that hold all the functionalities of the Paystack API.
    """

    @decorators.func_type_checker
    def __init__(self, secret_key: str) -> None:
        self.secret_key = secret_key

    def __repr__(self):
        return 'Paystack(paystack_secret_key)'

    def transaction(self):
        return transactions.Transaction(self.secret_key)

    def transactionsplit(self):
        return transactions_split.TransactionSplit(self.secret_key)

    def customers(self):
        return customers.Customer(self.secret_key)

    def dedicated_virtual_accounts(self):
        return virtual_accounts.DedicatedVirtualAccounts(self.secret_key)

    def applepay(self):
        return apple_pay.ApplePay(self.secret_key)

    def subaccounts(self):
        return subaccount.SubAccounts(self.secret_key)

    def plans(self):
        return plans.Plan(self.secret_key)

    def subscriptions(self):
        return subscription.Subscription(self.secret_key)

    def product(self):
        return products.Product(self.secret_key)

    def payment_pages(self):
        return payment_pages.PaymentPages(self.secret_key)

    def invoices(self):
        return invoices.Invoice(self.secret_key)

    def settlement(self):
        return settlement.Settlements(self.secret_key)

    def transfer_recipient(self):
        return transfer_recipient.TransferRecipient(self.secret_key)

    def transfer(self):
        return transfers.Tranfer(self.secret_key)

    def transfer_control(self):
        return transfers_control.TransferControl(self.secret_key)

    def bulk_charges(self):
        return bulk_charges.BulkCharges(self.secret_key)

    def control_panel(self):
        return control_panel.ControlPanel(self.secret_key)

    def charge(self):
        return charge.Charge(self.secret_key)

    def dispute(self):
        return disputes.Disputes(self.secret_key)

    def refund(self):
        return refund.Refund(self.secret_key)

    def verification(self):
        return verification.Verification(self.secret_key)

    def miscellaneous(self):
        return miscellaneous.Miscellaneous(self.secret_key)



class Error(Exception):
    """Base class for other exceptions"""

class UnwantedArgumentsError(Error):
    """raised when a dealing with unwanted arguments
    """

class MissingArgumentsError(Error):
    """raised when an argument is missing
    """



import datetime
import json

from ..utilities import settings, util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class Transaction(Request):

    """
    Accept payment from your customers.
    The Transactions API allows you create and manage payments on your integration
    """

    path = '/transaction'

    def initialize(self, email: str, amount: int, callback_url: str, reference: str = None, currency: str = None, plan: str = None, invoice_limit: int = None, channels: str | list = None, split_code: str = None, transaction_charge: int = None, subaccount: str = None, bearer: str = 'account', metadata: dict = None, generate_reference: bool = False):
        """Initialize a transaction from your backend

        Args:
            email (str): Customer's email address
            amount (int): Amount should be in kobo if currency is NGN, pesewas,
                if currency is GHS, and cents, if currency is ZAR
            callback_url (str): Fully qualified url, e.g. https://example.com/ .
                Use this to override the callback url provided on the dashboard for this transaction
            reference (str, optional): Unique transaction reference.
                Only -, ., = and alphanumeric characters allowed. Defaults to None.
            currency (str, optional): The transaction currency (NGN, GHS, ZAR or USD).
                Defaults to your integration currency.. Defaults to None.
            plan (str, optional): If transaction is to create a subscription to a predefined plan, provide plan code here.
                This would invalidate the value provided in amount. Defaults to None.
            invoice_limit (int, optional): Number of times to charge customer during subscription to plan. Defaults to None.
            channels (str | list, optional): An array of payment channels ( or a string of one channel ) to control what channels you want to make available to the user to make a payment with.
                Available channels include: ['card', 'bank', 'ussd', 'qr', 'mobile_money', 'bank_transfer']. Defaults to None.
            split_code (str, optional): The split code of the transaction split. e.g. SPL_98WF13Eb3w. Defaults to None.
            transaction_charge (int, optional): An amount used to override the split configuration for a single split payment.
                If set, the amount specified goes to the main account regardless of the split configuration. Defaults to None.
            subaccount (str, optional): The code for the subaccount that owns the payment.
                e.g. ACCT_8f4s1eq7ml6rlzj. Defaults to None.
            bearer (str, optional): Who bears Paystack charges? account or subaccount (defaults to account).
            metadata (dict, optional): Stringified JSON object of custom data.
                Kindly check the Metadata page for more information.. Defaults to None.
            generate_reference (bool, optional): Would you want a reference generated for you.
                It will override the reference if provided. Defaults to False.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/initialize'

        payload = {
            'amount': amount,
            'email': util.check_email(email),
            'callback_url': callback_url
        }

        if invoice_limit:
            payload['invoice_limit'] = invoice_limit

        if currency:
            payload['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        if reference:
            payload['reference'] = reference

        if generate_reference:
            payload['reference'] = util.create_ref()

        if plan:
            payload.pop('amount')
            payload['plan'] = util.check_code(
                settings.PLAN, plan)

        if channels:
            payload['channels'] = util.check_channels(channels)

        if split_code:
            payload['split_code'] = util.check_code(
                settings.SPLIT, split_code)

        if transaction_charge:
            payload['transaction_charge'] = transaction_charge

        if subaccount:
            payload['subaccount'] = util.check_code(
                settings.SUBACCOUNT, subaccount)
            if bearer:
                payload['bearer'] = util.check_membership(
                    settings.BEARER_TYPES, bearer, 'bearer_type')

        if metadata:
            payload['metadata'] = json.dumps(metadata)

        return self.post(path, payload=payload)

    def verify(self, reference: str):
        """Confirm the status of a transaction

        Args:
            reference (str): The transaction reference used to intiate the transaction

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/verify/{reference}'
        return self.get(path)

    def list_transactions(self, per_page: int = None, page: int = None, customer: int = None, status: str = None, from_date: datetime.datetime | datetime.date | str = None, to_date: datetime.datetime | datetime.date | str = None, amount: int = None):
        """List transactions carried out on your integration.

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.. Defaults to None.
            page (int, optional): Specify exactly what page you want to retrieve.
                If not specify we use a default value of 1.
            customer (int, optional): Specify an ID for the customer whose transactions you want to retrieve. Defaults to None.
            status (str, optional): Filter transactions by status ('failed', 'success', 'abandoned'). Defaults to None.
            from_date (datetime.datetime | datetime.date | str, optional): A timestamp from which to start listing transaction
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            to_date (datetime.datetime | datetime.date | str, optional): A timestamp at which to stop listing transaction
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            amount (int, optional): Filter transactions by amount. Specify the amount 
                (in kobo if currency is NGN, pesewas, if currency is GHS, and cents, if currency is ZAR). Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.generate_payload(
            locals(), 'per_page', 'page', 'status', 'from_date', 'to_date')
        params.update(util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date))

        if status:
            params['status'] = util.check_membership(
                settings.TRANSACTION_STATUS, status, 'status')

        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch(self, transaction_id: int):
        """Get details of a transaction carried out on your integration.


        Args:
            transaction_id (int): ID of the transaction to fetch.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{transaction_id}'
        return self.get(path)

    def check_authorization(self, email: str, amount: int, authorization_code: str, currency: str = None):
        """This feature is only available to businesses in Nigeria.
            All Mastercard and Visa authorizations can be checked with this endpoint to know
            if they have funds for the payment you seek.

            This endpoint should be used when you do not know the exact amount to charge a card when rendering a service.
            It should be used to check if a card has enough funds based on a maximum range value. It is well suited for:

            Ride hailing services
            Logistics services

            WARNING.
            You shouldn't use this endpoint to check a card for sufficient funds if you are going to charge the user immediately.
            This is because we hold funds when this endpoint is called which can lead to an insufficient funds error.

        Args:
            email (str): Customer's email address
            amount (int): Amount should be in kobo if currency is NGN, pesewas, if currency is GHS, and cents, if currency is ZAR
            authorization_code (str): Valid authorization code to charge
            currency (str, optional): Currency in which amount should be charged. Allowed values are: NGN, GHS, ZAR or USD. Defaults to None.

        Returns:
            JSON: Data fetched from API            
        """

        path = f'{self.path}/check_authorization'

        payload = {
            'email': util.check_email(email),
            'amount': amount,
            'authorization_code': util.check_code(settings.AUTHORIZATION, authorization_code)
        }

        if currency:
            payload['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        return self.post(path, payload=payload)

    def charge_authorization(self, email: str, amount: int, authorization_code: str, currency: str = None, reference: str = None, channels: list | str = None, queue: bool = False, split_code: str = None, subaccount: str = None, transaction_charge: int = None, bearer: str = None, metadata: dict = None, generate_reference: bool = False):
        """All authorizations marked as reusable can be charged with this endpoint whenever you need to receive payments.

        Args:
            email (str): Customer's email address
            amount (int): Amount should be in kobo if currency is NGN, pesewas,
                if currency is GHS, and cents, if currency is ZAR
            authorization_code (str): Valid authorization code to charge
            currency (str, optional): Currency in which amount should be charged.
                Allowed values are: NGN, GHS, ZAR or USD. Defaults to None.
            reference (str, optional): Unique transaction reference.
                Only -, ., = and alphanumeric characters allowed.. Defaults to None.
            channels (list | str, optional): Send us 'card' or 'bank' or 'card','bank' as an array
                ( or just a string if you plan to use one payment option )
                to specify what options to show the user paying. Defaults to None.
            queue (bool, optional): If you are making a scheduled charge call,
                it is a good idea to queue them so the processing system does not get overloaded causing transaction processing errors.
                Send queue:true to take advantage of our queued charging. Defaults to False.
            split_code (str, optional): _description_. Defaults to None.
            subaccount (str, optional): The code for the subaccount that owns the payment.
                e.g. ACCT_8f4s1eq7ml6rlzj. Defaults to None.
            transaction_charge (int, optional): A flat fee to charge the subaccount for this transaction
                (in kobo if currency is NGN, pesewas, if currency is GHS, and cents, if currency is ZAR).
                This overrides the split percentage set when the subaccount was created.
                Ideally, you will need to use this if you are splitting in flat rates
                (since subaccount creation only allows for percentage split).
                e.g. 7000 for a 70 naira flat fee.. Defaults to None.
            bearer (str, optional): Who bears Paystack charges? account or subaccount (defaults to account)
            metadata (dict, optional): A dictionary with custom_fields attribute which has an array of objects
                if you would like the fields to be added to your transaction when displayed on the dashboard.
                Sample: {"custom_fields":[{"display_name":"Cart ID","variable_name": "cart_id","value": "8393"}]}. Defaults to None.
            generate_reference (bool, optional): Whether to generate a reference for you. Defaults to False.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/charge_authorization'

        payload = {
            'email': util.check_email(email),
            'amount': amount,
            'authorization_code': util.check_code(settings.AUTHORIZATION, authorization_code),
        }

        if currency:
            payload['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        if reference:
            payload['reference'] = reference

        if not reference and generate_reference:
            payload['reference'] = util.create_ref()

        if channels:
            payload.update(util.check_channels(channels))

        if queue:
            payload['queue'] = True

        if split_code:
            payload['split_code'] = util.check_code(
                settings.SPLIT, split_code)

        if subaccount:
            payload['subaccount'] = util.check_code(
                settings.SUBACCOUNT, subaccount)
            if bearer:
                payload['bearer'] = util.check_membership(
                    settings.BEARER_TYPES, bearer, 'bearer_type')

        if transaction_charge:
            payload['transaction_charge'] = transaction_charge

        return self.post(path, payload=payload)

    def partial_debit(self, email: str, amount: int, currency: str, authorization_code: str, reference: str = None, at_least: str = None, generate_reference: bool = False):
        """Retrieve part of a payment from a customer

        Args:
            email (str): Customer's email address
            amount (int): Amount should be in kobo if currency is NGN, pesewas, 
                if currency is GHS, and cents, if currency is ZAR
            currency (str): Specify the currency you want to debit. Allowed values are NGN, GHS, ZAR or USD.
            authorization_code (str): Authorization Code
            reference (str, optional): Unique transaction reference. Only -, ., = and alphanumeric characters allowed. Defaults to None.
            at_least (str, optional): Minimum amount to charge. Defaults to None.
            generate_reference (bool, optional): Whether to generate reference for you. Defaults to False.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/part_debit'
        payload = {
            'email': util.check_email(email),
            'amount': amount,
            'currency': util.check_membership(settings.CURRENCIES, currency, 'currency'),
            'authorization_code': util.check_code(settings.AUTHORIZATION, authorization_code),
        }

        if reference:
            payload['reference'] = reference

        if not reference and generate_reference:
            payload['reference'] = util.create_ref()

        if at_least:
            assert at_least.isdigit(), "Invalid at_least value eg. '12000'"
            payload['at_least'] = at_least

        return self.post(path, payload=payload)

    def totals(self, per_page: int = None, page: int = None, from_date: datetime.datetime | datetime.date | str = None, to_date: datetime.datetime | datetime.date | str = None):
        """Total amount received on your account

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.
            page (int, optional): Specify exactly what page you want to retrieve.
                If not specify we use a default value of 1.
            from_date (datetime.datetime | datetime.date | str, optional): A timestamp from which to start listing transaction
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            to_date (datetime.datetime | datetime.date | str, optional): A timestamp at which to stop listing transaction
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/totals'
        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date).values()
        if params:
            return self.get(util.handle_query_params(path, params))
        return self.get(path)

    def timeline(self, id_or_reference: int | str = None):
        """View the timeline of a transaction.

        Args:
            id_or_reference (int | str, optional): The ID or the reference of the transaction. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f"{self.path}/timeline/{id_or_reference}" if id_or_reference else '/timeline'
        return self.get(path)

    def export(self, per_page: int = None, page: int = None, from_date: datetime.datetime | datetime.date | str = None, to_date: datetime.datetime | datetime.date | str = None, customer: int = None, status: str = None, currency: str = None, amount: int = None, settled: bool = None, settlement: int = None, payment_page: int = None):
        """List transactions carried out on your integration.

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.
            page (int, optional): Specify exactly what page you want to retrieve.
                If not specify we use a default value of 1.
            from_date (datetime.datetime | datetime.date | str, optional): A timestamp from which to start listing transaction
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            to_date (datetime.datetime | datetime.date | str, optional): A timestamp at which to stop listing transaction e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            customer (int, optional): Specify an ID for the customer whose transactions you want to retrieve. Defaults to None.
            status (str, optional): Filter transactions by status ('failed', 'success', 'abandoned'). Defaults to None.
            currency (str, optional): Specify the transaction currency to export.
                Allowed values are: in kobo if currency is NGN, pesewas, if currency is GHS, and cents, if currency is ZAR. Defaults to None.
            amount (int, optional): Filter transactions by amount. Specify the amount, in kobo
                if currency is NGN, pesewas, if currency is GHS, and cents, if currency is ZAR. Defaults to None.
            settled (bool, optional): Set to true to export only settled transactions. false for pending transactions.
                Leave undefined to export all transactions. Defaults to None.
            settlement (int, optional): An ID for the settlement whose transactions we should export. Defaults to None.
            payment_page (int, optional): Specify a payment page's id to export only transactions conducted on said page. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        params = util.generate_payload(
            locals(), 'per_page', 'page', 'from_date', 'to_date')
        params.update(util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date))

        if currency:
            params['currency'] = util.check_membership(
                settings.CURRENCIES, currency, 'currency')

        if status:
            params['status'] = util.check_membership(
                settings.TRANSACTION_STATUS, status, 'status')

        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)



import datetime

from ..utilities import settings, util, decorators
from ..utilities.request import Request


@decorators.class_type_checker
class SubAccounts(Request):

    """
    The Subaccounts API allows you create and manage subaccounts on your integration.
    Subaccounts can be used to split payment between two accounts (your main account and a sub account)
    """

    path = '/subaccount'

    def create(self, business_name: str, settlement_bank: str, account_number: str, percentage_charge: float, description: str = None, primary_contact_email: str = None, primary_contact_name: str = None, primary_contact_phone: str = None, metadata: str = None):
        """Create a subacount on your integration

        Args:
            business_name (str): Name of business for subaccount
            settlement_bank (str): Bank Code for the bank.
                You can get the list of Bank Codes by calling the List Banks endpoint.
            account_number (str): Bank Account Number
            percentage_charge (float): The default percentage charged when receiving on behalf of this subaccount
            description (str, optional): A description for this subaccount. Defaults to None.
            primary_contact_email (str, optional): A contact email for the subaccount. Defaults to None.
            primary_contact_name (str, optional): A name for the contact person for this subaccount. Defaults to None.
            primary_contact_phone (str, optional): A phone number to call for this subaccount. Defaults to None.
            metadata (str, optional): Stringified JSON object.
                Add a custom_fields attribute which has an array of objects if you would like the fields to be added to your transaction when displayed on the dashboard.
                Sample: {"custom_fields":[{"display_name":"Cart ID","variable_name": "cart_id","value": "8393"}]}.
                Defaults to None.

        Returns:
            JSON: Data fetched from API

        """

        payload = util.generate_payload(
            locals(), 'settlement_bank', 'account_number', 'primary_contact_email')
        payload['settlement_bank'] = util.check_bank_code(settlement_bank)
        payload['account_number'] = util.check_account_number(account_number)

        if primary_contact_email:
            payload['primary_contact_email'] = util.check_email(
                primary_contact_email)
        return self.post(self.path, payload=payload)

    def list_subaccounts(self, per_page: int = None, page: int = None, from_date: datetime.date | datetime.datetime | str = None, to_date: datetime.date | datetime.datetime | str = None):
        """List subaccounts available on your integration.

        Args:
            per_page (int, optional): Specify how many records you want to retrieve per page.
                If not specify we use a default value of 50.
            page (int, optional): Specify exactly what page you want to retrieve.
                If not specify we use a default value of 1.
            from_date (datetime.date | datetime.datetime | str, optional): A timestamp from which to start listing subaccounts
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.
            to_date (datetime.date | datetime.datetime | str, optional): A timestamp at which to stop listing subaccounts
                e.g. 2016-09-24T00:00:05.000Z, 2016-09-21. Defaults to None.

        Returns:
            JSON: Data fetched from API
        """
        params = util.check_query_params(
            per_page=per_page, page=page, from_date=from_date, to_date=to_date)
        if params:
            return self.get(util.handle_query_params(self.path, params))
        return self.get(self.path)

    def fetch(self, subaccount: int | str):
        """Get details of a subaccount on your integration. 

        Args:
            subaccount (int | str): The subaccount ID or code you want to fetch.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{util.id_or_code(subaccount, data=settings.SUBACCOUNT)}'

        return self.get(path)

    def update(self, subaccount: int | str, active: bool = None, business_name: str = None, settlement_bank: str = None, account_number: str = None, settlement_schedule: str = None, percentage_charge: float = None, description: str = None, primary_contact_email: str = None, primary_contact_name: str = None, primary_contact_phone: str = None, metadata: str = None):
        """Update a subaccount details on your integration.

        Args:
            subaccount (int | str): Subaccount's ID or code.
            active (bool, optional): Activate or deactivate a subaccount.
                Set value to true to activate subaccount or false to deactivate the subaccount.
            business_name (str, optional): Name of business for subaccount. Defaults to None.
            settlement_bank (str, optional): Bank Code for the bank.
                You can get the list of Bank Codes by calling the List Banks endpoint. Defaults to None.
            account_number (str, optional): Bank Account Number. Defaults to None.
            settlement_schedule (str, optional): Any of auto, weekly, `monthly`, `manual`.
                Auto means payout is T+1 and manual means payout to the subaccount should only be made when requested. Defaults to auto.
            percentage_charge (float, optional): The default percentage charged when receiving on behalf of this subaccount. Defaults to None.
            description (str, optional): A description for this subaccount. Defaults to None.
            primary_contact_email (str, optional): A contact email for the subaccount. Defaults to None.
            primary_contact_name (str, optional): A name for the contact person for this subaccount. Defaults to None.
            primary_contact_phone (str, optional): A phone number to call for this subaccount. Defaults to None.
            metadata (str, optional): Stringified JSON object. Add a custom_fields attribute which has an array of objects
                if you would like the fields to be added to your transaction when displayed on the dashboard.
                Sample: {"custom_fields":[{"display_name":"Cart ID","variable_name": "cart_id","value": "8393"}]}.
                Defaults to None.

        Returns:
            JSON: Data fetched from API
        """

        path = f'{self.path}/{util.id_or_code(subaccount, data=settings.SUBACCOUNT)}'

        payload = util.generate_payload(
            locals(), 'subaccount_id', 'subaccount_code')

        if settlement_bank:
            payload['settlement_bank'] = util.check_bank_code(settlement_bank)

        if settlement_schedule:
            payload['settlement_schedule'] = util.check_membership(
                settings.SETTLEMENT_SCHEDULES, settlement_schedule, 'settlement_schedule')

        if account_number:
            payload['account_number'] = util.check_account_number(
                account_number)

        if primary_contact_email:
            payload['primary_contact_email'] = util.check_email(
                primary_contact_email)

        return self.put(path, payload=payload)
